from dataclasses import dataclass, field
from typing import Dict, List, Tuple

import numpy as np
import numpy.typing as npt

from ludwig.api_annotations import DeveloperAPI, PublicAPI


@DeveloperAPI
@dataclass
class FeatureAttribution:
    """Stores the attribution for a single input feature."""

    # The name of the input feature.
    feature_name: str

    # The scalar attribution for the input feature.
    attribution: float

    # (Optional) The attribution for each token in the input feature as an array of shape (seq_len, 2).
    token_attributions: List[Tuple[str, float]] = None


@DeveloperAPI
@dataclass
class LabelExplanation:
    """Stores the feature attributions for a single label in the target feature's vocab."""

    # The attribution for each input feature.
    feature_attributions: List[FeatureAttribution] = field(default_factory=list)

    def add(self, feature_name: str, attribution: float, token_attributions: List[Tuple[str, float]] = None):
        """Add the attribution for a single input feature."""
        self.feature_attributions.append(FeatureAttribution(feature_name, attribution, token_attributions))

    def to_array(self) -> npt.NDArray[np.float64]:
        """Convert the explanation to a 1D array of shape (num_features,)."""
        return np.array([fa.attribution for fa in self.feature_attributions])


@DeveloperAPI
@dataclass
class Explanation:
    """Stores the explanations for a single row of input data.

    Contains the feature attributions for each label in the target feature's vocab.
    """

    target: str

    # The explanations for each label in the vocab of the target feature.
    label_explanations: List[LabelExplanation] = field(default_factory=list)

    def add(
        self,
        feat_names: List[str],
        feat_attributions: npt.NDArray[np.float64],
        feat_to_token_attributions: Dict[str, List[Tuple[str, float]]] = None,
        prepend: bool = False,
    ):
        """Add the feature attributions for a single label."""
        assert len(feat_names) == len(
            feat_attributions
        ), f"Expected {len(feat_names)} feature attributions, got {len(feat_attributions)}"
        if len(self.label_explanations) > 0:
            # Check that the feature attributions are the same shape as existing explanations.
            assert self.label_explanations[0].to_array().shape == feat_attributions.shape, (
                f"Expected feature attributions of shape {self.label_explanations[0].to_array().shape}, "
                f"got {feat_attributions.shape}"
            )

        le = LabelExplanation()
        for i, feat_name in enumerate(feat_names):
            le.add(
                feat_name,
                feat_attributions[i],
                feat_to_token_attributions.get(feat_name) if feat_to_token_attributions else None,
            )
        self.label_explanations.insert(0, le) if prepend else self.label_explanations.append(le)

    def to_array(self) -> npt.NDArray[np.float64]:
        """Convert the explanation to a 2D array of shape (num_labels, num_features)."""
        return np.array([le.to_array() for le in self.label_explanations])


@PublicAPI(stability="experimental")
@dataclass
class ExplanationsResult:
    # Aggregate explanation for the entire input data.
    global_explanation: Explanation  # GlobalExplanation

    # A list of explanations, one for each row in the input data.
    # Each explanation contains the feature attributions for each label in the target feature's vocab.
    row_explanations: List[Explanation]

    # Expected value for each label in the target feature's vocab.
    expected_values: List[float]



from typing import Dict, List, Type, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.encoders.base import Encoder
from ludwig.utils.registry import Registry

_encoder_registry = Registry()
_sequence_encoder_registry = Registry()


@DeveloperAPI
def get_encoder_registry() -> Registry:
    return _encoder_registry


@DeveloperAPI
def get_sequence_encoder_registry() -> Registry:
    return _sequence_encoder_registry


def register_sequence_encoder(name: str):
    def wrap(cls):
        get_sequence_encoder_registry()[name] = cls
        return cls

    return wrap


def register_encoder(name: str, features: Union[str, List[str]]):
    if isinstance(features, str):
        features = [features]

    def update_registry(registry_getter_fn, cls, feature):
        feature_registry = registry_getter_fn().get(feature, {})
        feature_registry[name] = cls
        registry_getter_fn()[feature] = feature_registry

    def wrap(cls):
        for feature in features:
            update_registry(get_encoder_registry, cls, feature)
        return cls

    return wrap


def get_encoder_cls(feature: str, name: str) -> Type[Encoder]:
    return get_encoder_registry()[feature][name]


def get_encoder_classes(feature: str) -> Dict[str, Type[Encoder]]:
    return get_encoder_registry()[feature]



"""Ludwig: Data-centric declarative deep learning framework."""
from codecs import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the README.md file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    requirements = [line.strip() for line in f if line]

extra_requirements = {}

with open(path.join(here, "requirements_serve.txt"), encoding="utf-8") as f:
    extra_requirements["serve"] = [line.strip() for line in f if line]

with open(path.join(here, "requirements_viz.txt"), encoding="utf-8") as f:
    extra_requirements["viz"] = [line.strip() for line in f if line]

with open(path.join(here, "requirements_distributed.txt"), encoding="utf-8") as f:
    extra_requirements["distributed"] = [line.strip() for line in f if line]

with open(path.join(here, "requirements_hyperopt.txt"), encoding="utf-8") as f:
    extra_requirements["hyperopt"] = [line.strip() for line in f if line]

with open(path.join(here, "requirements_tree.txt"), encoding="utf-8") as f:
    extra_requirements["tree"] = [line.strip() for line in f if line]

with open(path.join(here, "requirements_explain.txt"), encoding="utf-8") as f:
    extra_requirements["explain"] = [line.strip() for line in f if line]

with open(path.join(here, "requirements_benchmarking.txt"), encoding="utf-8") as f:
    extra_requirements["benchmarking"] = [line.strip() for line in f if line]

extra_requirements["full"] = [item for sublist in extra_requirements.values() for item in sublist]

with open(path.join(here, "requirements_test.txt"), encoding="utf-8") as f:
    extra_requirements["test"] = extra_requirements["full"] + [line.strip() for line in f if line]

setup(
    name="ludwig",
    version="0.7.3",
    description="Declarative machine learning: End-to-end machine learning pipelines using data-driven configurations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ludwig-ai/ludwig",
    download_url="https://pypi.org/project/ludwig/",
    author="Piero Molino",
    author_email="piero.molino@gmail.com",
    license="Apache 2.0",
    keywords="ludwig deep learning deep_learning machine machine_learning natural language processing computer vision",
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    python_requires=">=3.8",
    include_package_data=True,
    package_data={"ludwig": ["etc/*", "examples/*.py"]},
    install_requires=requirements,
    extras_require=extra_requirements,
    entry_points={"console_scripts": ["ludwig=ludwig.cli:main"]},
)



from typing import List, Tuple, Union

import pandas as pd
from sklearn.model_selection import train_test_split

from ludwig.api_annotations import PublicAPI
from ludwig.constants import TEST_SPLIT, TRAIN_SPLIT, VALIDATION_SPLIT
from ludwig.data.dataset.base import Dataset
from ludwig.utils.defaults import default_random_seed


@PublicAPI
def get_repeatable_train_val_test_split(
    df_input, stratify_colname="", random_seed=default_random_seed, frac_train=0.7, frac_val=0.1, frac_test=0.2
):
    """Return df_input with split column containing (if possible) non-zero rows in the train, validation, and test
    data subset categories.

    If the input dataframe does not contain an existing split column or if the
    number of rows in both the validation and test split is 0 and non-empty
    stratify_colname specified, return df_input with split column set according
    to frac_<subset_name> and stratify_colname.

    Else stratify_colname is ignored, and:
     If the input dataframe contains an existing split column and non-zero row
      counts for all three split types, return df_input.
     If the input dataframe contains an existing split column but only one of
      validation and test split has non-zero row counts, return df_input with
      missing split getting rows from train split as per frac_<subset_name>.

    Parameters
    ----------
    df_input : Pandas dataframe
        Input dataframe to be split.
    stratify_colname : str
        The column used for stratification (if desired); usually the label column.
    random_seed : int
        Seed used to get repeatable split.
    frac_train : float
    frac_val   : float
    frac_test  : float
        The ratios with which to split the dataframe into train, val, and test data;
        should sum to 1.0.

    Returns
    -------
    df_split :
        Dataframe containing the three splits.
    """

    if frac_train + frac_val + frac_test != 1.0:
        raise ValueError(f"fractions {frac_train:f}, {frac_val:f}, {frac_test:f} do not add up to 1.0")
    if stratify_colname:
        do_stratify_split = True
        if stratify_colname not in df_input.columns:
            raise ValueError("%s is not a column in the dataframe" % (stratify_colname))
    else:
        do_stratify_split = False
        if "split" not in df_input.columns:
            df_input["split"] = 0  # set up for non-stratified split path

    if "split" in df_input.columns:
        df_train = df_input[df_input["split"] == TRAIN_SPLIT].copy()
        df_val = df_input[df_input["split"] == VALIDATION_SPLIT].copy()
        df_test = df_input[df_input["split"] == TEST_SPLIT].copy()
        if not do_stratify_split or len(df_val) != 0 or len(df_test) != 0:
            if len(df_val) == 0:
                df_val = df_train.sample(frac=frac_val, replace=False, random_state=random_seed)
                df_train = df_train.drop(df_val.index)
            if len(df_test) == 0:
                df_test = df_train.sample(frac=frac_test, replace=False, random_state=random_seed)
                df_train = df_train.drop(df_test.index)
            do_stratify_split = False

    if do_stratify_split:
        # Make sure the `stratify_colname` doesn't have any NaNs.
        df_input = df_input[df_input[stratify_colname].notna()]

        # Split original dataframe into train and temp dataframes.
        y = df_input[[stratify_colname]]  # Dataframe of just the column on which to stratify.
        df_train, df_temp, y_train, y_temp = train_test_split(
            df_input, y, stratify=y, test_size=(1.0 - frac_train), random_state=random_seed
        )
        # Split the temp dataframe into val and test dataframes.
        relative_frac_test = frac_test / (frac_val + frac_test)
        df_val, df_test, y_val, y_test = train_test_split(
            df_temp, y_temp, stratify=y_temp, test_size=relative_frac_test, random_state=random_seed
        )

    assert len(df_input) == len(df_train) + len(df_val) + len(df_test)
    df_train["split"] = TRAIN_SPLIT
    df_val["split"] = VALIDATION_SPLIT
    df_test["split"] = TEST_SPLIT
    df_split = pd.concat([df_train, df_val, df_test], ignore_index=True)
    return df_split


def generate_dataset_statistics(
    training_set: Dataset, validation_set: Union[Dataset, None], test_set: Union[Dataset, None]
) -> List[Tuple[str, int, int]]:
    from ludwig.benchmarking.utils import format_memory

    dataset_statistics = [["Dataset", "Size (Rows)", "Size (In Memory)"]]
    dataset_statistics.append(["Training", len(training_set), format_memory(training_set.in_memory_size_bytes)])
    if validation_set is not None:
        dataset_statistics.append(
            ["Validation", len(validation_set), format_memory(validation_set.in_memory_size_bytes)]
        )
    if test_set is not None:
        dataset_statistics.append(["Test", len(test_set), format_memory(test_set.in_memory_size_bytes)])
    return dataset_statistics



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import functools
import logging
from io import BytesIO
from typing import Any, List, Optional, Union

import torch
import torch.nn.functional as F
import torchaudio

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import DEFAULT_AUDIO_TENSOR_LENGTH
from ludwig.utils.fs_utils import get_bytes_obj_from_path
from ludwig.utils.types import TorchAudioTuple

logger = logging.getLogger(__name__)

# https://github.com/pytorch/audio/blob/main/torchaudio/csrc/sox/types.cpp
AUDIO_EXTENSIONS = (".wav", ".amb", ".mp3", ".ogg", ".vorbis", ".flac", ".opus", ".sphere")


@DeveloperAPI
def is_torch_audio_tuple(audio: Any) -> bool:
    if isinstance(audio, tuple):
        if len(audio) == 2 and isinstance(audio[0], torch.Tensor) and isinstance(audio[1], int):
            return True
    return False


@DeveloperAPI
def get_default_audio(audio_lst: List[TorchAudioTuple]) -> TorchAudioTuple:
    sampling_rates = [audio[1] for audio in audio_lst]
    tensor_list = [audio[0] for audio in audio_lst]

    for i, tensor in enumerate(tensor_list):
        if tensor.shape[1] > DEFAULT_AUDIO_TENSOR_LENGTH:
            tensor_list[i] = tensor[:, :DEFAULT_AUDIO_TENSOR_LENGTH]
        else:
            pad_size = DEFAULT_AUDIO_TENSOR_LENGTH - tensor.shape[1]
            tensor_list[i] = F.pad(tensor, (0, pad_size))
    default_audio_tensor = torch.mean(torch.stack(tensor_list), dim=0)
    default_sampling_rate = calculate_mean(sum(sampling_rates), len(sampling_rates))

    return default_audio_tensor, default_sampling_rate


@DeveloperAPI
def read_audio_from_path(path: str) -> Optional[TorchAudioTuple]:
    """Reads audio from path.

    Useful for reading from a small number of paths. For more intensive reads, use backend.read_binary_files instead.
    """
    bytes_obj = get_bytes_obj_from_path(path)
    return read_audio_from_bytes_obj(bytes_obj)


@DeveloperAPI
@functools.lru_cache(maxsize=32)
def read_audio_from_bytes_obj(bytes_obj: bytes) -> Optional[TorchAudioTuple]:
    try:
        f = BytesIO(bytes_obj)
        return torchaudio.backend.sox_io_backend.load(f)
    except Exception as e:
        logger.warning(e)
        return None


def _pre_emphasize_data(data: torch.Tensor, emphasize_value: float = 0.97):
    # Increase precision in order to achieve parity with scipy.signal.lfilter implementation
    filter_window = torch.tensor([1.0, -emphasize_value], dtype=torch.float64, device=data.device)
    a_coeffs = torch.tensor([1, 0], dtype=torch.float64, device=data.device)
    pre_emphasized_data = torchaudio.functional.lfilter(
        data.to(dtype=torch.float64),
        a_coeffs,
        filter_window,
        clamp=False,
    ).to(torch.float32)
    return pre_emphasized_data


@DeveloperAPI
def get_length_in_samp(sampling_rate_in_hz: Union[float, int], length_in_s: Union[float, int]) -> int:
    return int(sampling_rate_in_hz * length_in_s)


@DeveloperAPI
def get_group_delay(
    raw_data: torch.Tensor,
    sampling_rate_in_hz: int,
    window_length_in_s: float,
    window_shift_in_s: float,
    num_fft_points: int,
    window_type: str,
):
    X_stft_transform = _get_stft(
        raw_data, sampling_rate_in_hz, window_length_in_s, window_shift_in_s, num_fft_points, window_type=window_type
    )
    Y_stft_transform = _get_stft(
        raw_data,
        sampling_rate_in_hz,
        window_length_in_s,
        window_shift_in_s,
        num_fft_points,
        window_type=window_type,
        data_transformation="group_delay",
    )
    X_stft_transform_real = torch.real(X_stft_transform)
    X_stft_transform_imag = torch.imag(X_stft_transform)
    Y_stft_transform_real = torch.real(Y_stft_transform)
    Y_stft_transform_imag = torch.imag(Y_stft_transform)
    nominator = torch.multiply(X_stft_transform_real, Y_stft_transform_real) + torch.multiply(
        X_stft_transform_imag, Y_stft_transform_imag
    )
    denominator = torch.square(torch.abs(X_stft_transform))
    group_delay = torch.divide(nominator, denominator + 1e-10)
    assert not torch.isnan(group_delay).any(), "There are NaN values in group delay"
    return torch.transpose(group_delay, 0, 1)


@DeveloperAPI
def get_phase_stft_magnitude(
    raw_data: torch.Tensor,
    sampling_rate_in_hz: int,
    window_length_in_s: float,
    window_shift_in_s: float,
    num_fft_points: int,
    window_type: str,
) -> torch.Tensor:
    stft = _get_stft(
        raw_data, sampling_rate_in_hz, window_length_in_s, window_shift_in_s, num_fft_points, window_type=window_type
    )
    abs_stft = torch.abs(stft)
    phase = torch.angle(stft)
    stft_phase = torch.cat([phase, abs_stft], dim=1)
    return torch.transpose(stft_phase, 0, 1)


@DeveloperAPI
def get_stft_magnitude(
    raw_data: torch.Tensor,
    sampling_rate_in_hz: int,
    window_length_in_s: float,
    window_shift_in_s: float,
    num_fft_points: int,
    window_type: str,
):
    stft = _get_stft(
        raw_data, sampling_rate_in_hz, window_length_in_s, window_shift_in_s, num_fft_points, window_type=window_type
    )
    stft_magnitude = torch.abs(stft)
    return torch.transpose(stft_magnitude, 0, 1)


################################################################################
# The following code for FBank is adapted from jameslyons/python_speech_features
# MIT licensed implementation
# https://github.com/jameslyons/python_speech_features/blob/40c590269b57c64a8c1f1ddaaff2162008d1850c/python_speech_features/base.py#L84################################################################################
################################################################################
@DeveloperAPI
def get_fbank(
    raw_data: torch.Tensor,
    sampling_rate_in_hz: int,
    window_length_in_s: float,
    window_shift_in_s: float,
    num_fft_points: int,
    window_type: str,
    num_filter_bands: int,
) -> torch.Tensor:
    stft = _get_stft(
        raw_data,
        sampling_rate_in_hz,
        window_length_in_s,
        window_shift_in_s,
        num_fft_points,
        window_type=window_type,
        zero_mean_offset=True,
    )
    stft_power = torch.abs(stft) ** 2
    upper_limit_freq = int(sampling_rate_in_hz / 2)
    upper_limit_mel = _convert_hz_to_mel(upper_limit_freq)
    lower_limit_mel = 0
    list_mel_points = torch.linspace(lower_limit_mel, upper_limit_mel, num_filter_bands + 2, device=raw_data.device)
    mel_fbank_matrix = _get_mel_fbank_matrix(list_mel_points, num_filter_bands, num_fft_points, sampling_rate_in_hz)
    mel_fbank_feature = torch.matmul(stft_power, torch.transpose(mel_fbank_matrix, 0, 1))
    log_mel_fbank_feature = torch.log(mel_fbank_feature + 1.0e-10)
    return torch.transpose(log_mel_fbank_feature, 0, 1)


def _get_mel_fbank_matrix(
    list_mel_points: torch.Tensor, num_filter_bands: int, num_fft_points: int, sampling_rate_in_hz: int
) -> torch.Tensor:
    num_ess_fft_points = get_non_symmetric_length(num_fft_points)
    freq_scale = (num_fft_points + 1) / sampling_rate_in_hz
    freq_bins_on_mel_scale = torch.floor(freq_scale * _convert_mel_to_hz(list_mel_points))
    mel_scaled_fbank = torch.zeros(
        (num_filter_bands, num_ess_fft_points), dtype=torch.float32, device=list_mel_points.device
    )
    for filt_idx in range(num_filter_bands):
        start_bin_freq = freq_bins_on_mel_scale[filt_idx]
        middle_bin_freq = freq_bins_on_mel_scale[filt_idx + 1]
        end_bin_freq = freq_bins_on_mel_scale[filt_idx + 2]
        mel_scaled_fbank[filt_idx] = _create_triangular_filter(
            start_bin_freq, middle_bin_freq, end_bin_freq, num_ess_fft_points
        )
    return mel_scaled_fbank


def _create_triangular_filter(
    start_bin_freq: torch.Tensor, middle_bin_freq: torch.Tensor, end_bin_freq: torch.Tensor, num_ess_fft_points: int
):
    filter_window = torch.zeros(num_ess_fft_points, dtype=torch.float32, device=start_bin_freq.device)
    filt_support_begin = middle_bin_freq - start_bin_freq
    filt_support_end = end_bin_freq - middle_bin_freq
    for freq in range(int(start_bin_freq), int(middle_bin_freq)):
        filter_window[freq] = (freq - start_bin_freq) / filt_support_begin
    for freq in range(int(middle_bin_freq), int(end_bin_freq)):
        filter_window[freq] = (end_bin_freq - freq) / filt_support_end
    return filter_window


def _convert_hz_to_mel(hz: int) -> float:
    return float(2595.0 * torch.log10(torch.tensor(1 + hz / 700.0)))


def _convert_mel_to_hz(mel):
    return 700.0 * (10 ** (mel / 2595.0) - 1)


def _get_stft(
    raw_data: torch.Tensor,
    sampling_rate_in_hz: int,
    window_length_in_s: float,
    window_shift_in_s: float,
    num_fft_points: int,
    window_type: str,
    data_transformation: Optional[str] = None,
    zero_mean_offset: bool = False,
) -> torch.Tensor:
    pre_emphasized_data = _pre_emphasize_data(raw_data)
    stft = _short_time_fourier_transform(
        pre_emphasized_data,
        sampling_rate_in_hz,
        window_length_in_s,
        window_shift_in_s,
        num_fft_points,
        window_type,
        data_transformation,
        zero_mean_offset,
    )
    non_symmetric_stft = get_non_symmetric_data(stft)
    return non_symmetric_stft


def _short_time_fourier_transform(
    data: torch.Tensor,
    sampling_rate_in_hz: int,
    window_length_in_s: float,
    window_shift_in_s: float,
    num_fft_points: int,
    window_type: str,
    data_transformation: Optional[str] = None,
    zero_mean_offset: bool = False,
) -> torch.Tensor:
    window_length_in_samp: int = get_length_in_samp(window_length_in_s, sampling_rate_in_hz)
    window_shift_in_samp: int = get_length_in_samp(window_shift_in_s, sampling_rate_in_hz)
    preprocessed_data_matrix = _preprocess_to_padded_matrix(
        data[0], window_length_in_samp, window_shift_in_samp, zero_mean_offset=zero_mean_offset
    )
    weighted_data_matrix = _weight_data_matrix(
        preprocessed_data_matrix, window_type, data_transformation=data_transformation
    )
    fft = torch.fft.fft(weighted_data_matrix, n=num_fft_points)
    return fft


def _preprocess_to_padded_matrix(
    data: torch.Tensor, window_length_in_samp: int, window_shift_in_samp: int, zero_mean_offset: bool = False
) -> torch.Tensor:
    num_input = data.shape[0]
    num_output = get_num_output_padded_to_fit_input(num_input, window_length_in_samp, window_shift_in_samp)
    zero_padded_matrix = torch.zeros((num_output, window_length_in_samp), dtype=torch.float32, device=data.device)
    for num_output_idx in range(num_output):
        start_idx = window_shift_in_samp * num_output_idx
        is_last_output = num_output_idx == num_output - 1
        end_idx = start_idx + window_length_in_samp if not is_last_output else num_input
        end_padded_idx = window_length_in_samp if not is_last_output else end_idx - start_idx
        window_data = data[start_idx:end_idx]
        if zero_mean_offset:
            window_data = window_data - torch.mean(window_data)
        zero_padded_matrix[num_output_idx, :end_padded_idx] = window_data
    return zero_padded_matrix


@DeveloperAPI
def get_num_output_padded_to_fit_input(num_input: int, window_length_in_samp: int, window_shift_in_samp: int) -> int:
    num_output_valid = torch.tensor((num_input - window_length_in_samp) / window_shift_in_samp + 1)
    return int(torch.ceil(num_output_valid))


@DeveloperAPI
def get_window(window_type: str, window_length_in_samp: int, device: Optional[torch.device] = None) -> torch.Tensor:
    # Increase precision in order to achieve parity with scipy.signal.windows.get_window implementation
    if window_type == "bartlett":
        return torch.bartlett_window(window_length_in_samp, periodic=False, dtype=torch.float64, device=device).to(
            torch.float32
        )
    elif window_type == "blackman":
        return torch.blackman_window(window_length_in_samp, periodic=False, dtype=torch.float64, device=device).to(
            torch.float32
        )
    elif window_type == "hamming":
        return torch.hamming_window(window_length_in_samp, periodic=False, dtype=torch.float64, device=device).to(
            torch.float32
        )
    elif window_type == "hann":
        return torch.hann_window(window_length_in_samp, periodic=False, dtype=torch.float64, device=device).to(
            torch.float32
        )
    else:
        raise ValueError(f"Unknown window type: {window_type}")


@DeveloperAPI
def is_audio_score(src_path):
    # Used for AutoML
    return int(isinstance(src_path, str) and src_path.lower().endswith(AUDIO_EXTENSIONS))


def _weight_data_matrix(
    data_matrix: torch.Tensor, window_type: str, data_transformation: Optional[str] = None
) -> torch.Tensor:
    window_length_in_samp = data_matrix[0].shape[0]
    window = get_window(window_type, window_length_in_samp, device=data_matrix.device)
    if data_transformation is not None and data_transformation == "group_delay":
        window *= torch.arange(window_length_in_samp, device=data_matrix.device).float()
    return data_matrix * window


@DeveloperAPI
def get_non_symmetric_length(symmetric_length: int) -> int:
    return int(symmetric_length / 2) + 1


@DeveloperAPI
def get_non_symmetric_data(data: torch.Tensor) -> torch.Tensor:
    num_fft_points = data.shape[-1]
    num_ess_fft_points = get_non_symmetric_length(num_fft_points)
    return data[:, :num_ess_fft_points]


@DeveloperAPI
def get_max_length_stft_based(length_in_samp, window_length_in_s, window_shift_in_s, sampling_rate_in_hz):
    window_length_in_samp = get_length_in_samp(window_length_in_s, sampling_rate_in_hz)
    window_shift_in_samp = get_length_in_samp(window_shift_in_s, sampling_rate_in_hz)
    return get_num_output_padded_to_fit_input(length_in_samp, window_length_in_samp, window_shift_in_samp)


@DeveloperAPI
def calculate_incr_var(var_prev, mean_prev, mean, length):
    return var_prev + (length - mean_prev) * (length - mean)


@DeveloperAPI
def calculate_incr_mean(count, mean, length):
    return mean + (length - mean) / float(count)


@DeveloperAPI
def calculate_var(sum1, sum2, count):
    return (sum2 - ((sum1 * sum1) / float(count))) / float(count - 1) if count > 1 else 0.0


@DeveloperAPI
def calculate_mean(sum1, count):
    return sum1 / float(count)



# Register all encoder schemas
import ludwig.schema.encoders.bag_encoders
import ludwig.schema.encoders.category_encoders
import ludwig.schema.encoders.date_encoders
import ludwig.schema.encoders.h3_encoders
import ludwig.schema.encoders.image
import ludwig.schema.encoders.sequence_encoders
import ludwig.schema.encoders.set_encoders
import ludwig.schema.encoders.text_encoders  # noqa



from collections import OrderedDict

import torch

from ludwig.utils import metric_utils
from ludwig.utils.metric_utils import TrainerMetric


def test_dynamic_partition():
    data = torch.Tensor([10, 20, 30, 40, 50])
    partitions = torch.Tensor([0, 0, 1, 1, 0])

    partitioned_data = metric_utils.dynamic_partition(data, partitions, 2)

    assert torch.equal(partitioned_data[0], torch.Tensor([10.0, 20.0, 50.0]))
    assert torch.equal(partitioned_data[1], torch.Tensor([30.0, 40.0]))


def test_dynamic_partition_2D():
    data = torch.Tensor(
        [
            [1, 2, 3, 4, 5, 6, 7, 8, 9],
            [10, 11, 12, 13, 14, 15, 16, 17, 18],
        ]
    )
    partitions = torch.Tensor([[1, 1, 1, 1, 1, 1, 1, 1, 0], [1, 1, 1, 1, 1, 1, 1, 1, 0]])

    partitioned_data = metric_utils.dynamic_partition(data, partitions, 2)

    assert torch.equal(partitioned_data[0], torch.Tensor([9, 18]))
    assert torch.equal(
        partitioned_data[1],
        torch.Tensor([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0]),
    )


def test_masked_correct_predictions():
    preds = torch.tensor([[1, 5, 1, 5, 1, 5, 12, 12, 12], [10, 1, 5, 1, 5, 12, 12, 12, 12]])
    targets = torch.tensor([[1, 9, 5, 7, 5, 9, 13, 6, 0], [1, 9, 7, 13, 4, 7, 7, 7, 0]])
    targets_sequence_length = torch.tensor([8, 8])

    result = metric_utils.masked_correct_predictions(targets, preds, targets_sequence_length)

    assert torch.equal(
        result, torch.Tensor([1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    )


def test_reduce_trainer_metrics_dict():
    dict_dict_trainer_metrics = {
        "feature_name": {"metric_name": [metric_utils.TrainerMetric(epoch=1, step=2, value=10)]}
    }

    result = metric_utils.reduce_trainer_metrics_dict(dict_dict_trainer_metrics)

    assert result == {"feature_name": {"metric_name": [10]}}


def test_reduce_trainer_metrics_dict_ordered_dict():
    dict_dict_trainer_metrics = OrderedDict(
        [
            (
                "category_5B6BF",
                OrderedDict(
                    [
                        ("loss", [TrainerMetric(epoch=0, step=1, value=0.0)]),
                        ("accuracy", [TrainerMetric(epoch=0, step=1, value=1.0)]),
                    ]
                ),
            ),
            ("combined", {"loss": [TrainerMetric(epoch=0, step=1, value=0.0)]}),
        ]
    )

    result = metric_utils.reduce_trainer_metrics_dict(dict_dict_trainer_metrics)

    assert result == {"category_5B6BF": {"accuracy": [1.0], "loss": [0.0]}, "combined": {"loss": [0.0]}}



import json
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Union

from dataclasses_json import dataclass_json

from ludwig.api_annotations import DeveloperAPI
from ludwig.utils.misc_utils import memoized_method


@DeveloperAPI
class ExpectedImpact(int, Enum):
    """The expected impact of determining a "good" value for a specific parameter.

    - HIGH: this parameter should almost always be included in a hyperopt run and can make or break a good model.
    - MEDIUM: this parameter can sometimes make or break a good model.
    - LOW: this parameter usually does not have a significant impact on model performance.
    """

    UNKNOWN = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


@DeveloperAPI
@dataclass_json()
@dataclass
class ParameterMetadata:
    """Contains descriptive information that pertains to a Ludwig configuration parameter."""

    # Quick description generally for UI display
    short_description: str = ""

    # In depth description generally for documentation purposes.
    long_description: str = ""

    # How this parameter can be displayed in a human-readable form.
    ui_display_name: Union[str, None] = ""

    # Why the default value for this parameter is the default.
    default_value_reasoning: Union[str, None] = None

    # Examples of other values that can be used for this parameter.
    example_value: Union[List[Any], None] = None

    # List of related parameters that this parameter interacts with or depends on.
    related_parameters: Union[List[str], None] = None

    # Other information that is relevant for this parameter.
    other_information: Union[str, None] = None

    # If we change/increase/decrease this parameter, what's the intuition for how model performance would change, i.e.:
    # learning curves, model speed, memory usage, etc.
    description_implications: Union[str, None] = None

    # What values would a machine learning expert suggest users try to help improve their model? Ideally, covers 95%
    # (~2 sigma) of use cases.
    suggested_values: Any = None

    # The reasoning behind the suggested values, as well as model performance indicators or other intuition that could
    # help inform a user to make an educated decision about what values to experiment with for this parameter.
    suggested_values_reasoning: Union[str, None] = None

    # True if this parameter could be frequently used, would have a high impact, and/or would be interesting for a
    # machine learning practitioner.
    commonly_used: bool = False

    # The expected impact of determining a "good" value for this parameter.
    expected_impact: ExpectedImpact = ExpectedImpact.UNKNOWN

    # List of links, papers, and blog posts to learn more.
    literature_references: Union[List[str], None] = None

    # Whether the parameter is used strictly internally.
    internal_only: bool = False

    @memoized_method(maxsize=1)
    def to_json_dict(self) -> Dict[str, Any]:
        return json.loads(self.to_json())


@DeveloperAPI
def convert_metadata_to_json(pm: ParameterMetadata) -> Dict[str, Any]:
    """Converts a ParameterMetadata dict to a normal JSON dict.

    NOTE: Without the json.loads call, to_json() returns
    a string repr that is improperly parsed.
    """
    if not pm:
        return None
    return pm.to_json_dict()


# This is a quick way to flag schema parameters as internal only via the `parameter_metadata` argument
INTERNAL_ONLY = ParameterMetadata(internal_only=True)



"""Utilities used for managing output feature dicts."""

from typing import Dict, List

import numpy as np
import torch

from ludwig.utils.torch_utils import sequence_length_3D, sequence_mask


def get_feature_concat_name(feature_name: str, tensor_name: str) -> str:
    return feature_name + "::" + tensor_name


def get_tensor_name_from_concat_name(concat_name: str) -> str:
    return concat_name.split("::")[-1]


def get_feature_name_from_concat_name(concat_name: str) -> str:
    return "::".join(concat_name.split("::")[:-1])


def get_single_output_feature_tensors(
    output_feature_dict: Dict[str, torch.Tensor], feature_name: str
) -> Dict[str, torch.Tensor]:
    """Returns a map of tensors related to the given feature_name."""
    single_output_feature_tensors = {}
    for concat_name, tensor in output_feature_dict.items():
        if get_feature_name_from_concat_name(concat_name) == feature_name:
            single_output_feature_tensors[get_tensor_name_from_concat_name(concat_name)] = tensor
    return single_output_feature_tensors


def get_output_feature_tensor(
    output_dict: Dict[str, torch.Tensor], feature_name: str, tensor_name: str
) -> torch.Tensor:
    """Returns a tensor related for the given feature_name and tensor_name."""
    concat_name = get_feature_concat_name(feature_name, tensor_name)
    if concat_name not in output_dict:
        raise ValueError(
            f"Could not find {tensor_name} for {feature_name} in the output_dict with keys: {output_dict.keys()}"
        )
    return output_dict[get_feature_concat_name(feature_name, tensor_name)]


def set_output_feature_tensor(
    output_dict: Dict[str, torch.Tensor], feature_name: str, tensor_name: str, tensor: torch.Tensor
):
    """Adds tensor for the given feature_name and tensor_name to the tensor dict."""
    output_dict[get_feature_concat_name(feature_name, tensor_name)] = tensor


def concat_dependencies(
    feature_name: str,
    dependencies: List[str],
    dependency_reducers: torch.ModuleDict,
    combiner_hidden_state: torch.Tensor,
    other_output_feature_states: Dict[str, torch.Tensor],
) -> torch.Tensor:
    """Concatenates combiner_hidden_state with other output feature hidden states based on listed dependencies."""
    # No dependencies.
    if not dependencies:
        return combiner_hidden_state

    dependency_hidden_states = []
    for feature_name in dependencies:
        # The dependent feature should be present since ECD does a topological sort over output features.
        feature_hidden_state = other_output_feature_states[feature_name]

        # This feature is sequential.
        if len(combiner_hidden_state.shape) > 2:
            if len(feature_hidden_state.shape) > 2:
                # The dependent feature is also sequential.
                # matrix matrix -> concat
                assert combiner_hidden_state.shape[1] == feature_hidden_state.shape[1]
                dependency_hidden_states.append(feature_hidden_state)
            else:
                # The dependent feature is not sequential.
                # matrix vector -> tile concat
                sequence_max_length = combiner_hidden_state.shape[1]
                multipliers = (1, sequence_max_length, 1)
                tiled_representation = torch.tile(torch.unsqueeze(feature_hidden_state, 1), multipliers)

                sequence_length = sequence_length_3D(combiner_hidden_state)
                mask = sequence_mask(sequence_length, sequence_max_length)
                tiled_representation = torch.mul(
                    tiled_representation,
                    mask[:, :, np.newaxis].type(torch.float32),
                )

                dependency_hidden_states.append(tiled_representation)

        else:
            # This feature is not sequential.
            if len(feature_hidden_state.shape) > 2:
                # The dependent feature is sequential.
                # vector matrix -> reduce concat
                reducer = dependency_reducers[feature_name]
                dependency_hidden_states.append(reducer(feature_hidden_state))
            else:
                # The dependent feature is not sequential.
                # vector vector -> concat
                dependency_hidden_states.append(feature_hidden_state)

    try:
        hidden = torch.cat([combiner_hidden_state] + dependency_hidden_states, dim=-1)
    except Exception as e:
        raise ValueError(
            f"Shape mismatch {e} while concatenating dependent features of {feature_name}: "
            f"{dependencies}. Concatenating the feature activations tensor {combiner_hidden_state} "
            f"with activation tensors of dependencies: {dependency_hidden_states}. The error is "
            "likely due to a mismatch of the second dimension (sequence length) or a "
            "difference in ranks. Likely solutions are setting the maximum_sequence_length "
            "of all sequential features to be the same,  or reduce the output of some "
            "features, or disabling the bucketing setting bucketing_field to None / null, "
            "as activating it will reduce the length of the field the bucketing is "
            "performed on."
        )
    return hidden



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import contextlib
import logging
import multiprocessing
import os
import random
import shutil
import sys
import tempfile
import traceback
import uuid
from distutils.util import strtobool
from typing import Any, Dict, List, Optional, Set, Tuple, TYPE_CHECKING, Union

import cloudpickle
import numpy as np
import pandas as pd
import pytest
import torch
from PIL import Image

from ludwig.api import LudwigModel
from ludwig.backend import LocalBackend
from ludwig.constants import (
    AUDIO,
    BAG,
    BATCH_SIZE,
    BINARY,
    CATEGORY,
    COLUMN,
    DATE,
    DECODER,
    ENCODER,
    H3,
    IMAGE,
    NAME,
    NUMBER,
    PROC_COLUMN,
    SEQUENCE,
    SET,
    SPLIT,
    TEXT,
    TIMESERIES,
    TRAINER,
    VECTOR,
)
from ludwig.data.dataset_synthesizer import build_synthetic_dataset, DATETIME_FORMATS
from ludwig.experiment import experiment_cli
from ludwig.features.feature_utils import compute_feature_hash
from ludwig.globals import PREDICTIONS_PARQUET_FILE_NAME
from ludwig.trainers.trainer import Trainer
from ludwig.utils import fs_utils
from ludwig.utils.data_utils import read_csv, replace_file_extension, use_credentials

if TYPE_CHECKING:
    from ludwig.data.dataset.base import Dataset
    from ludwig.schema.model_types.base import ModelConfig

logger = logging.getLogger(__name__)

# Used in sequence-related unit tests (encoders, features) as well as end-to-end integration tests.
# Missing: passthrough encoder.
ENCODERS = ["embed", "rnn", "parallel_cnn", "cnnrnn", "stacked_parallel_cnn", "stacked_cnn", "transformer"]

HF_ENCODERS_SHORT = ["distilbert"]

HF_ENCODERS = [
    "bert",
    "gpt",
    "gpt2",
    "transformer_xl",
    "xlnet",
    # "xlm",  # disabled in the schema: https://github.com/ludwig-ai/ludwig/pull/3108
    "roberta",
    "distilbert",
    # "ctrl",  # disabled in the schema: https://github.com/ludwig-ai/ludwig/pull/2976
    "camembert",
    "albert",
    "t5",
    "xlmroberta",
    "longformer",
    "flaubert",
    "electra",
    # "mt5",    # disabled in the schema: https://github.com/ludwig-ai/ludwig/pull/2982
]

RAY_BACKEND_CONFIG = {
    "type": "ray",
    "processor": {
        "parallelism": 2,
    },
    "trainer": {
        "use_gpu": False,
        "num_workers": 1,
        "resources_per_worker": {
            "CPU": 0.1,
            "GPU": 0,
        },
    },
}


class LocalTestBackend(LocalBackend):
    @property
    def supports_multiprocessing(self):
        return False


# Simulates running training on a separate node from the driver process
class FakeRemoteBackend(LocalBackend):
    def create_trainer(self, **kwargs) -> "BaseTrainer":  # noqa: F821
        return FakeRemoteTrainer(**kwargs)

    @property
    def supports_multiprocessing(self):
        return False


class FakeRemoteTrainer(Trainer):
    def train(self, *args, save_path="model", **kwargs):
        with tempfile.TemporaryDirectory() as tmpdir:
            return super().train(*args, save_path=tmpdir, **kwargs)


def parse_flag_from_env(key, default=False):
    try:
        value = os.environ[key]
    except KeyError:
        # KEY isn't set, default to `default`.
        _value = default
    else:
        # KEY is set, convert it to True or False.
        try:
            _value = strtobool(value)
        except ValueError:
            # More values are supported, but let's keep the message simple.
            raise ValueError(f"If set, {key} must be yes or no.")
    return _value


_run_private_tests = parse_flag_from_env("RUN_PRIVATE", default=False)


def private_param(param):
    """Wrap param to mark it as private, meaning it requires credentials to run.

    Private tests are skipped by default. Set the RUN_PRIVATE environment variable to a truth value to run them.
    """
    return pytest.param(
        *param,
        marks=pytest.mark.skipif(
            not _run_private_tests,
            reason="Skipping: this test is marked private, set RUN_PRIVATE=1 in your environment to run",
        ),
    )


def generate_data(
    input_features,
    output_features,
    filename="test_csv.csv",
    num_examples=25,
    nan_percent=0.0,
):
    """Helper method to generate synthetic data based on input, output feature specs.

    :param num_examples: number of examples to generate
    :param input_features: schema
    :param output_features: schema
    :param filename: path to the file where data is stored
    :param nan_percent: percent of values in a feature to be NaN
    :return:
    """
    df = generate_data_as_dataframe(input_features, output_features, num_examples, nan_percent)
    df.to_csv(filename, index=False)
    return filename


def generate_data_as_dataframe(
    input_features,
    output_features,
    num_examples=25,
    nan_percent=0.0,
) -> pd.DataFrame:
    """Helper method to generate synthetic data based on input, output feature specs.

    Args:
        num_examples: number of examples to generate
        input_features: schema
        output_features: schema
        nan_percent: percent of values in a feature to be NaN
    Returns:
        A pandas DataFrame
    """
    features = input_features + output_features
    df = build_synthetic_dataset(num_examples, features)
    data = [next(df) for _ in range(num_examples + 1)]

    return pd.DataFrame(data[1:], columns=data[0])


def recursive_update(dictionary, values):
    for k, v in values.items():
        if isinstance(v, dict):
            dictionary[k] = recursive_update(dictionary.get(k, {}), v)
        else:
            dictionary[k] = v
    return dictionary


def random_string(length=5):
    return uuid.uuid4().hex[:length].upper()


def number_feature(normalization=None, **kwargs):
    feature = {
        "name": f"{NUMBER}_{random_string()}",
        "type": NUMBER,
        "preprocessing": {"normalization": normalization},
    }
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def category_feature(output_feature=False, **kwargs):
    if DECODER in kwargs:
        output_feature = True
    feature = {
        "name": f"{CATEGORY}_{random_string()}",
        "type": CATEGORY,
    }
    if output_feature:
        feature.update(
            {
                DECODER: {"type": "classifier", "vocab_size": 10},
            }
        )
    else:
        feature.update(
            {
                ENCODER: {"vocab_size": 10, "embedding_size": 5},
            }
        )
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def text_feature(output_feature=False, **kwargs):
    if DECODER in kwargs:
        output_feature = True
    feature = {
        "name": f"{TEXT}_{random_string()}",
        "type": TEXT,
    }
    if output_feature:
        feature.update(
            {
                DECODER: {"type": "generator", "vocab_size": 5, "max_len": 7},
            }
        )
    else:
        feature.update(
            {
                ENCODER: {
                    "type": "parallel_cnn",
                    "vocab_size": 5,
                    "min_len": 7,
                    "max_len": 7,
                    "embedding_size": 8,
                    "state_size": 8,
                },
            }
        )
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def set_feature(output_feature=False, **kwargs):
    if DECODER in kwargs:
        output_feature = True
    feature = {
        "name": f"{SET}_{random_string()}",
        "type": SET,
    }
    if output_feature:
        feature.update(
            {
                DECODER: {"type": "classifier", "vocab_size": 10, "max_len": 5},
            }
        )
    else:
        feature.update(
            {
                ENCODER: {"type": "embed", "vocab_size": 10, "max_len": 5, "embedding_size": 5},
            }
        )
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def sequence_feature(output_feature=False, **kwargs):
    if DECODER in kwargs:
        output_feature = True
    feature = {
        "name": f"{SEQUENCE}_{random_string()}",
        "type": SEQUENCE,
    }
    if output_feature:
        feature.update(
            {
                DECODER: {
                    "type": "generator",
                    "vocab_size": 10,
                    "max_len": 7,
                }
            }
        )
    else:
        feature.update(
            {
                ENCODER: {
                    "type": "embed",
                    "vocab_size": 10,
                    "max_len": 7,
                    "embedding_size": 8,
                    "output_size": 8,
                    "state_size": 8,
                    "num_filters": 8,
                    "hidden_size": 8,
                },
            }
        )
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def image_feature(folder, **kwargs):
    feature = {
        "name": f"{IMAGE}_{random_string()}",
        "type": IMAGE,
        "preprocessing": {"in_memory": True, "height": 12, "width": 12, "num_channels": 3},
        ENCODER: {
            "type": "stacked_cnn",
        },
        "destination_folder": folder,
    }
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def audio_feature(folder, **kwargs):
    feature = {
        "name": f"{AUDIO}_{random_string()}",
        "type": AUDIO,
        "preprocessing": {
            "type": "fbank",
            "window_length_in_s": 0.04,
            "window_shift_in_s": 0.02,
            "num_filter_bands": 80,
            "audio_file_length_limit_in_s": 3.0,
        },
        ENCODER: {
            "type": "stacked_cnn",
            "should_embed": False,
            "conv_layers": [
                {"filter_size": 400, "pool_size": 16, "num_filters": 32},
                {"filter_size": 40, "pool_size": 10, "num_filters": 64},
            ],
            "output_size": 16,
        },
        "destination_folder": folder,
    }
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def timeseries_feature(**kwargs):
    feature = {
        "name": f"{TIMESERIES}_{random_string()}",
        "type": TIMESERIES,
        ENCODER: {"type": "parallel_cnn", "max_len": 7},
    }
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def binary_feature(**kwargs):
    feature = {
        "name": f"{BINARY}_{random_string()}",
        "type": BINARY,
    }
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def bag_feature(**kwargs):
    feature = {
        "name": f"{BAG}_{random_string()}",
        "type": BAG,
        ENCODER: {"type": "embed", "max_len": 5, "vocab_size": 10, "embedding_size": 5},
    }
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def date_feature(**kwargs):
    feature = {
        "name": f"{DATE}_{random_string()}",
        "type": DATE,
        "preprocessing": {
            "datetime_format": random.choice(list(DATETIME_FORMATS.keys())),
        },
    }
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def h3_feature(**kwargs):
    feature = {
        "name": f"{H3}_{random_string()}",
        "type": H3,
    }
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def vector_feature(**kwargs):
    feature = {
        "name": f"{VECTOR}_{random_string()}",
        "type": VECTOR,
        "preprocessing": {
            "vector_size": 5,
        },
    }
    recursive_update(feature, kwargs)
    feature[COLUMN] = feature[NAME]
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


def run_experiment(
    input_features=None, output_features=None, config=None, skip_save_processed_input=True, backend=None, **kwargs
):
    """Helper method to avoid code repetition in running an experiment. Deletes the data saved to disk related to
    running an experiment.

    :param input_features: list of input feature dictionaries
    :param output_features: list of output feature dictionaries
    **kwargs you may also pass extra parameters to the experiment as keyword
    arguments
    :return: None
    """
    if input_features is None and output_features is None and config is None:
        raise ValueError("Cannot run test experiment without features nor config.")

    if config is None:
        config = {
            "input_features": input_features,
            "output_features": output_features,
            "combiner": {"type": "concat", "output_size": 14},
            TRAINER: {"epochs": 2, BATCH_SIZE: 128},
        }

    with tempfile.TemporaryDirectory() as tmpdir:
        args = {
            "config": config,
            "backend": backend or LocalTestBackend(),
            "skip_save_training_description": True,
            "skip_save_training_statistics": True,
            "skip_save_processed_input": skip_save_processed_input,
            "skip_save_progress": True,
            "skip_save_unprocessed_output": True,
            "skip_save_model": True,
            "skip_save_predictions": True,
            "skip_save_eval_stats": True,
            "skip_collect_predictions": True,
            "skip_collect_overall_stats": True,
            "skip_save_log": True,
            "output_directory": tmpdir,
        }
        args.update(kwargs)

        experiment_cli(**args)


def generate_output_features_with_dependencies(main_feature, dependencies):
    """Generates multiple output features specifications with dependencies.

    Example usage:
        generate_output_features_with_dependencies('sequence_feature', ['category_feature', 'number_feature'])

    Args:
        main_feature: feature identifier, valid values 'category_feature', 'sequence_feature', 'number_feature'
        dependencies: list of dependencies for 'main_feature', do not li
    """

    output_features = [
        category_feature(decoder={"type": "classifier", "vocab_size": 2}, reduce_input="sum", output_feature=True),
        sequence_feature(decoder={"type": "generator", "vocab_size": 10, "max_len": 5}, output_feature=True),
        number_feature(),
    ]

    # value portion of dictionary is a tuple: (position, feature_name)
    #   position: location of output feature in the above output_features list
    #   feature_name: Ludwig generated feature name
    feature_names = {
        "category_feature": (0, output_features[0]["name"]),
        "sequence_feature": (1, output_features[1]["name"]),
        "number_feature": (2, output_features[2]["name"]),
    }

    # generate list of dependencies with real feature names
    generated_dependencies = [feature_names[feat_name][1] for feat_name in dependencies]

    # specify dependencies for the main_feature
    output_features[feature_names[main_feature][0]]["dependencies"] = generated_dependencies

    return output_features


def generate_output_features_with_dependencies_complex():
    """Generates multiple output features specifications with dependencies."""

    tf = text_feature(decoder={"vocab_size": 4, "max_len": 5, "type": "generator"})
    sf = sequence_feature(decoder={"vocab_size": 4, "max_len": 5, "type": "generator"}, dependencies=[tf["name"]])
    nf = number_feature(dependencies=[tf["name"]])
    vf = vector_feature(dependencies=[sf["name"], nf["name"]])
    set_f = set_feature(decoder={"type": "classifier", "vocab_size": 4}, dependencies=[tf["name"], vf["name"]])
    cf = category_feature(
        decoder={"type": "classifier", "vocab_size": 4}, dependencies=[sf["name"], nf["name"], set_f["name"]]
    )

    # The correct order ids[tf, sf, nf, vf, set_f, cf]
    # shuffling it to test the robustness of the topological sort
    output_features = [nf, tf, set_f, vf, cf, sf]

    return output_features


def _subproc_wrapper(fn, queue, *args, **kwargs):
    fn = cloudpickle.loads(fn)
    try:
        results = fn(*args, **kwargs)
    except Exception as e:
        traceback.print_exc(file=sys.stderr)
        results = e
    queue.put(results)


def spawn(fn):
    def wrapped_fn(*args, **kwargs):
        ctx = multiprocessing.get_context("spawn")
        queue = ctx.Queue()

        p = ctx.Process(target=_subproc_wrapper, args=(cloudpickle.dumps(fn), queue, *args), kwargs=kwargs)

        p.start()
        p.join()
        results = queue.get()
        if isinstance(results, Exception):
            raise RuntimeError(
                f"Spawned subprocess raised {type(results).__name__}, " f"check log output above for stack trace."
            )
        return results

    return wrapped_fn


def get_weights(model: torch.nn.Module) -> List[torch.Tensor]:
    return [param.data for param in model.parameters()]


def has_no_grad(
    val: Union[np.ndarray, torch.Tensor, str, list],
):
    """Checks if two values are close to each other."""
    if isinstance(val, list):
        return all(has_no_grad(v) for v in val)
    if isinstance(val, torch.Tensor):
        return not val.requires_grad
    return True


def is_all_close(
    val1: Union[np.ndarray, torch.Tensor, str, list],
    val2: Union[np.ndarray, torch.Tensor, str, list],
    tolerance=1e-4,
):
    """Checks if two values are close to each other."""
    if isinstance(val1, list):
        return all(is_all_close(v1, v2, tolerance) for v1, v2 in zip(val1, val2))
    if isinstance(val1, str):
        return val1 == val2
    if isinstance(val1, torch.Tensor):
        val1 = val1.cpu().detach().numpy()
    if isinstance(val2, torch.Tensor):
        val2 = val2.cpu().detach().numpy()
    return val1.shape == val2.shape and np.allclose(val1, val2, atol=tolerance)


def is_all_tensors_cuda(val: Union[np.ndarray, torch.Tensor, str, list]) -> bool:
    if isinstance(val, list):
        return all(is_all_tensors_cuda(v) for v in val)

    if isinstance(val, torch.Tensor):
        return val.is_cuda
    return True


def run_api_experiment(input_features, output_features, data_csv):
    """Helper method to avoid code repetition in running an experiment.

    :param input_features: input schema
    :param output_features: output schema
    :param data_csv: path to data
    :return: None
    """
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    model = LudwigModel(config)
    output_dir = None

    try:
        # Training with csv
        _, _, output_dir = model.train(
            dataset=data_csv, skip_save_processed_input=True, skip_save_progress=True, skip_save_unprocessed_output=True
        )
        model.predict(dataset=data_csv)

        model_dir = os.path.join(output_dir, "model")
        loaded_model = LudwigModel.load(model_dir)

        # Necessary before call to get_weights() to materialize the weights
        loaded_model.predict(dataset=data_csv)

        model_weights = get_weights(model.model)
        loaded_weights = get_weights(loaded_model.model)
        for model_weight, loaded_weight in zip(model_weights, loaded_weights):
            assert torch.allclose(model_weight, loaded_weight)
    finally:
        # Remove results/intermediate data saved to disk
        shutil.rmtree(output_dir, ignore_errors=True)

    try:
        # Training with dataframe
        data_df = read_csv(data_csv)
        _, _, output_dir = model.train(
            dataset=data_df, skip_save_processed_input=True, skip_save_progress=True, skip_save_unprocessed_output=True
        )
        model.predict(dataset=data_df)
    finally:
        shutil.rmtree(output_dir, ignore_errors=True)


def add_nans_to_df_in_place(df: pd.DataFrame, nan_percent: float):
    """Adds nans to a pandas dataframe in-place."""
    if nan_percent == 0:
        # No-op if nan_percent is 0
        return None
    if nan_percent < 0 or nan_percent > 1:
        raise ValueError("nan_percent must be between 0 and 1")

    num_rows = len(df)
    num_nans_per_col = int(round(nan_percent * num_rows))
    for col in df.columns:
        if col == SPLIT:  # do not add NaNs to the split column
            continue
        col_idx = df.columns.get_loc(col)
        for row_idx in random.sample(range(num_rows), num_nans_per_col):
            df.iloc[row_idx, col_idx] = np.nan
    return None


def read_csv_with_nan(path, nan_percent=0.0):
    """Converts `nan_percent` of samples in each row of the CSV at `path` to NaNs."""
    df = pd.read_csv(path)
    add_nans_to_df_in_place(df, nan_percent)
    return df


def create_data_set_to_use(data_format, raw_data, nan_percent=0.0):
    # helper function for generating training and test data with specified format
    # handles all data formats except for hdf5
    # assumes raw_data is a csv dataset generated by
    # tests.integration_tests.utils.generate_data() function

    # support for writing to a fwf dataset based on this stackoverflow posting:
    # https://stackoverflow.com/questions/16490261/python-pandas-write-dataframe-to-fixed-width-file-to-fwf
    from tabulate import tabulate

    def to_fwf(df, fname):
        content = tabulate(df.values.tolist(), list(df.columns), tablefmt="plain")
        open(fname, "w").write(content)

    pd.DataFrame.to_fwf = to_fwf

    dataset_to_use = None

    if data_format == "csv":
        # Replace the original CSV with a CSV with NaNs
        dataset_to_use = raw_data
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_csv(dataset_to_use, index=False)

    elif data_format in {"df", "dict"}:
        dataset_to_use = read_csv_with_nan(raw_data, nan_percent=nan_percent)
        if data_format == "dict":
            dataset_to_use = dataset_to_use.to_dict(orient="list")

    elif data_format == "excel":
        dataset_to_use = replace_file_extension(raw_data, "xlsx")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_excel(dataset_to_use, index=False)

    elif data_format == "excel_xls":
        dataset_to_use = replace_file_extension(raw_data, "xls")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_excel(dataset_to_use, index=False)

    elif data_format == "feather":
        dataset_to_use = replace_file_extension(raw_data, "feather")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_feather(dataset_to_use)

    elif data_format == "fwf":
        dataset_to_use = replace_file_extension(raw_data, "fwf")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_fwf(dataset_to_use)

    elif data_format == "html":
        dataset_to_use = replace_file_extension(raw_data, "html")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_html(dataset_to_use, index=False)

    elif data_format == "json":
        dataset_to_use = replace_file_extension(raw_data, "json")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_json(dataset_to_use, orient="records")

    elif data_format == "jsonl":
        dataset_to_use = replace_file_extension(raw_data, "jsonl")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_json(dataset_to_use, orient="records", lines=True)

    elif data_format == "parquet":
        dataset_to_use = replace_file_extension(raw_data, "parquet")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_parquet(dataset_to_use, index=False)

    elif data_format == "pickle":
        dataset_to_use = replace_file_extension(raw_data, "pickle")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_pickle(dataset_to_use)

    elif data_format == "stata":
        dataset_to_use = replace_file_extension(raw_data, "stata")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_stata(dataset_to_use)

    elif data_format == "tsv":
        dataset_to_use = replace_file_extension(raw_data, "tsv")
        read_csv_with_nan(raw_data, nan_percent=nan_percent).to_csv(dataset_to_use, sep="\t", index=False)

    elif data_format == "pandas+numpy_images":
        df = read_csv_with_nan(raw_data, nan_percent=nan_percent)
        processed_df_rows = []
        for _, row in df.iterrows():
            processed_df_row = {}
            for feature_name, raw_feature in row.iteritems():
                if "image" in feature_name and not (type(raw_feature) == float and np.isnan(raw_feature)):
                    feature = np.array(Image.open(raw_feature))
                else:
                    feature = raw_feature
                processed_df_row[feature_name] = feature
            processed_df_rows.append(processed_df_row)
        dataset_to_use = pd.DataFrame(processed_df_rows)

    else:
        ValueError(f"'{data_format}' is an unrecognized data format")

    return dataset_to_use


def augment_dataset_with_none(
    df: pd.DataFrame, first_row_none: bool = False, last_row_none: bool = False, nan_cols: Optional[List] = None
) -> pd.DataFrame:
    """Optionally sets the first and last rows of nan_cols of the given dataframe to nan.

    :param df: dataframe containg input features/output features
    :type df: pd.DataFrame
    :param first_row_none: indicates whether to set the first rowin the dataframe to np.nan
    :type first_row_none: bool
    :param last_row_none: indicates whether to set the last row in the dataframe to np.nan
    :type last_row_none: bool
    :param nan_cols: a list of columns in the dataframe to explicitly set the first or last rows to np.nan
    :type nan_cols: list
    """
    nan_cols = nan_cols if nan_cols is not None else []

    if first_row_none:
        for col in nan_cols:
            df.iloc[0, df.columns.get_loc(col)] = np.nan
    if last_row_none:
        for col in nan_cols:
            df.iloc[-1, df.columns.get_loc(col)] = np.nan
    return df


def train_with_backend(
    backend,
    config,
    dataset=None,
    training_set=None,
    validation_set=None,
    test_set=None,
    predict=True,
    evaluate=True,
    callbacks=None,
    skip_save_processed_input=True,
    skip_save_predictions=True,
    required_metrics=None,
):
    model = LudwigModel(config, backend=backend, callbacks=callbacks)
    with tempfile.TemporaryDirectory() as output_directory:
        _, _, _ = model.train(
            dataset=dataset,
            training_set=training_set,
            validation_set=validation_set,
            test_set=test_set,
            skip_save_processed_input=skip_save_processed_input,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
            skip_save_log=True,
            output_directory=output_directory,
        )

        if dataset is None:
            dataset = training_set

        if predict:
            preds, _ = model.predict(
                dataset=dataset, skip_save_predictions=skip_save_predictions, output_directory=output_directory
            )
            assert preds is not None

            if not skip_save_predictions:
                read_preds = model.backend.df_engine.read_predictions(
                    os.path.join(output_directory, PREDICTIONS_PARQUET_FILE_NAME)
                )
                # call compute to ensure preds materialize correctly
                read_preds = read_preds.compute()
                assert read_preds is not None

        if evaluate:
            eval_stats, eval_preds, _ = model.evaluate(
                dataset=dataset, collect_overall_stats=False, collect_predictions=True
            )
            assert eval_preds is not None
            assert_all_required_metrics_exist(eval_stats, required_metrics)

            # Test that eval_stats are approx equal when using local backend
            with tempfile.TemporaryDirectory() as tmpdir:
                model.save(tmpdir)
                local_model = LudwigModel.load(tmpdir, backend=LocalTestBackend())
                local_eval_stats, _, _ = local_model.evaluate(
                    dataset=dataset, collect_overall_stats=False, collect_predictions=False
                )

                # Filter out metrics that are not being aggregated correctly for now
                # TODO(travis): https://github.com/ludwig-ai/ludwig/issues/1956
                def filter(stats):
                    return {
                        k: {
                            metric_name: value
                            for metric_name, value in v.items()
                            if metric_name
                            not in {"loss", "root_mean_squared_percentage_error", "jaccard", "token_accuracy"}
                        }
                        for k, v in stats.items()
                    }

                for (feature_name_from_eval, metrics_dict_from_eval), (
                    feature_name_from_local,
                    metrics_dict_from_local,
                ) in zip(filter(eval_stats).items(), filter(local_eval_stats).items()):
                    for (metric_name_from_eval, metric_value_from_eval), (
                        metric_name_from_local,
                        metric_value_from_local,
                    ) in zip(metrics_dict_from_eval.items(), metrics_dict_from_local.items()):
                        assert metric_name_from_eval == metric_name_from_local, (
                            f"Metric mismatch between eval and local. Metrics from eval: "
                            f"{metrics_dict_from_eval.keys()}. Metrics from local: {metrics_dict_from_local.keys()}"
                        )
                        assert np.isclose(metric_value_from_eval, metric_value_from_local, rtol=1e-03, atol=1e-04), (
                            f"Metric {metric_name_from_eval} for feature {feature_name_from_eval}: "
                            f"{metric_value_from_eval} != {metric_value_from_local}"
                        )

        return model


def assert_all_required_metrics_exist(
    feature_to_metrics_dict: Dict[str, Dict[str, Any]], required_metrics: Optional[Dict[str, Set]] = None
):
    """Checks that all `required_metrics` exist in the dictionary returned during Ludwig model evaluation.

    `feature_to_metrics_dict` is a dict where the feature name is a key and the value is a dictionary of metrics:

        {
            "binary_1234": {
                "accuracy": 0.5,
                "loss": 0.5,
            },
            "numerical_1234": {
                "mean_squared_error": 0.5,
                "loss": 0.5,
            }
        }

    `required_metrics` is a dict where the feature name is a key and the value is a set of metric names:

        {
            "binary_1234": {"accuracy"},
            "numerical_1234": {"mean_squared_error"},
        }

    Args:
        feature_to_metrics_dict: dictionary of output feature to a dictionary of metrics
        required_metrics: optional dictionary of output feature to a set of metrics names. If None, then function
            returns True immediately.
    Returns:
        None. Raises an AssertionError if any required metrics are missing.
    """
    if required_metrics is None:
        return

    for feature_name, metrics_dict in feature_to_metrics_dict.items():
        if feature_name in required_metrics:
            required_metric_names = set(required_metrics[feature_name])
            metric_names = set(metrics_dict.keys())
            assert required_metric_names.issubset(
                metric_names
            ), f"required metrics {required_metric_names} not in metrics {metric_names} for feature {feature_name}"


def assert_preprocessed_dataset_shape_and_dtype_for_feature(
    feature_name: str,
    preprocessed_dataset: "Dataset",
    config_obj: "ModelConfig",
    expected_dtype: np.dtype,
    expected_shape: Tuple,
):
    """Asserts that the preprocessed dataset has the correct shape and dtype for a given feature type.

    Args:
        feature_name: the name of the feature to check
        preprocessed_dataset: the preprocessed dataset
        config_obj: the model config object
        expected_dtype: the expected dtype
        expected_shape: the expected shape
    Returns:
        None.
    Raises:
        AssertionError if the preprocessed dataset does not have the correct shape and dtype for the given feature type.
    """
    if_configs = [if_config for if_config in config_obj.input_features if if_config.name == feature_name]
    # fail fast if given `feature_name`` is not found or is not unique
    if len(if_configs) != 1:
        raise ValueError(f"feature_name {feature_name} found {len(if_configs)} times in config_obj")
    if_config = if_configs[0]

    if_config_proc_column = if_config.proc_column
    for result in [
        preprocessed_dataset.training_set,
        preprocessed_dataset.validation_set,
        preprocessed_dataset.test_set,
    ]:
        result_df = result.to_df()
        result_df_proc_col = result_df[if_config_proc_column]

        # Check that the proc col is of the correct dtype
        result_df_proc_col_dtypes = set(result_df_proc_col.map(lambda x: x.dtype))
        assert all(
            [expected_dtype == dtype for dtype in result_df_proc_col_dtypes]
        ), f"proc dtype should be {expected_dtype}, got the following set of values: {result_df_proc_col_dtypes}"

        # Check that the proc col is of the right dimensions
        result_df_proc_col_shapes = set(result_df_proc_col.map(lambda x: x.shape))
        assert all(
            expected_shape == shape for shape in result_df_proc_col_shapes
        ), f"proc shape should be {expected_shape}, got the following set of values: {result_df_proc_col_shapes}"


@contextlib.contextmanager
def remote_tmpdir(fs_protocol, bucket):
    if bucket is None:
        with tempfile.TemporaryDirectory() as tmpdir:
            yield f"{fs_protocol}://{tmpdir}"
        return

    prefix = f"tmp_{uuid.uuid4().hex}"
    tmpdir = f"{fs_protocol}://{bucket}/{prefix}"
    try:
        with use_credentials(minio_test_creds()):
            fs_utils.makedirs(f"{fs_protocol}://{bucket}", exist_ok=True)
        yield tmpdir
    finally:
        try:
            with use_credentials(minio_test_creds()):
                fs_utils.delete(tmpdir, recursive=True)
        except FileNotFoundError as e:
            logger.info(f"failed to delete remote tempdir, does not exist: {str(e)}")
            pass


def minio_test_creds():
    return {
        "s3": {
            "client_kwargs": {
                "endpoint_url": os.environ.get("LUDWIG_MINIO_ENDPOINT", "http://localhost:9000"),
                "aws_access_key_id": os.environ.get("LUDWIG_MINIO_ACCESS_KEY", "minio"),
                "aws_secret_access_key": os.environ.get("LUDWIG_MINIO_SECRET_KEY", "minio123"),
            }
        }
    }



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Any, Dict, List, Union

import numpy as np
import torch

from ludwig.constants import COLUMN, HIDDEN, LOGITS, NAME, PREDICTIONS, PROBABILITIES, PROC_COLUMN, SET
from ludwig.features.base_feature import BaseFeatureMixin, InputFeature, OutputFeature, PredictModule
from ludwig.features.feature_utils import set_str_to_idx
from ludwig.schema.features.set_feature import SetInputFeatureConfig, SetOutputFeatureConfig
from ludwig.types import (
    FeatureMetadataDict,
    FeaturePostProcessingOutputDict,
    PreprocessingConfigDict,
    TrainingSetMetadataDict,
)
from ludwig.utils import output_feature_utils
from ludwig.utils.strings_utils import create_vocabulary, UNKNOWN_SYMBOL
from ludwig.utils.tokenizers import get_tokenizer_from_registry, TORCHSCRIPT_COMPATIBLE_TOKENIZERS
from ludwig.utils.types import TorchscriptPreprocessingInput

logger = logging.getLogger(__name__)


class _SetPreprocessing(torch.nn.Module):
    """Torchscript-enabled version of preprocessing done by SetFeatureMixin.add_feature_data.

    If is_bag is true, forward returns a vector for each sample indicating counts of each token. Else, forward returns a
    multi-hot vector for each sample indicating presence of each token.
    """

    def __init__(self, metadata: TrainingSetMetadataDict, is_bag: bool = False):
        super().__init__()
        if metadata["preprocessing"]["tokenizer"] not in TORCHSCRIPT_COMPATIBLE_TOKENIZERS:
            raise ValueError(
                f"{metadata['preprocessing']['tokenizer']} is not supported by torchscript. Please use "
                f"one of {TORCHSCRIPT_COMPATIBLE_TOKENIZERS}."
            )

        self.lowercase = metadata["preprocessing"]["lowercase"]
        self.tokenizer = get_tokenizer_from_registry(metadata["preprocessing"]["tokenizer"])()
        self.vocab_size = metadata["vocab_size"]
        self.unknown_symbol = UNKNOWN_SYMBOL
        self.unit_to_id = metadata["str2idx"]
        self.is_bag = is_bag

    def forward(self, v: TorchscriptPreprocessingInput) -> torch.Tensor:
        """Takes a list of strings and returns a tensor of counts for each token."""
        if not torch.jit.isinstance(v, List[str]):
            raise ValueError(f"Unsupported input: {v}")

        if self.lowercase:
            sequences = [sequence.lower() for sequence in v]
        else:
            sequences = v

        unit_sequences = self.tokenizer(sequences)
        # refines type of unit_sequences from Any to List[List[str]]
        assert torch.jit.isinstance(unit_sequences, List[List[str]]), "unit_sequences is not a list of lists."

        set_matrix = torch.zeros(len(unit_sequences), self.vocab_size, dtype=torch.float32)
        for sample_idx, unit_sequence in enumerate(unit_sequences):
            sequence_length = len(unit_sequence)
            for i in range(sequence_length):
                curr_unit = unit_sequence[i]
                if curr_unit in self.unit_to_id:
                    curr_id = self.unit_to_id[curr_unit]
                else:
                    curr_id = self.unit_to_id[self.unknown_symbol]

                if self.is_bag:
                    set_matrix[sample_idx][curr_id] += 1
                else:
                    set_matrix[sample_idx][curr_id] = 1

        return set_matrix


class _SetPostprocessing(torch.nn.Module):
    """Torchscript-enabled version of postprocessing done by SetFeatureMixin.add_feature_data."""

    def __init__(self, metadata: TrainingSetMetadataDict):
        super().__init__()
        self.idx2str = {i: v for i, v in enumerate(metadata["idx2str"])}
        self.predictions_key = PREDICTIONS
        self.probabilities_key = PROBABILITIES
        self.unk = UNKNOWN_SYMBOL

    def forward(self, preds: Dict[str, torch.Tensor], feature_name: str) -> FeaturePostProcessingOutputDict:
        predictions = output_feature_utils.get_output_feature_tensor(preds, feature_name, self.predictions_key)
        probabilities = output_feature_utils.get_output_feature_tensor(preds, feature_name, self.probabilities_key)

        inv_preds: List[List[str]] = []
        filtered_probs: List[torch.Tensor] = []
        for sample_idx, sample in enumerate(predictions):
            sample_preds: List[str] = []
            pos_sample_idxs: List[int] = []
            pos_class_idxs: List[int] = []
            for class_idx, is_positive in enumerate(sample):
                if is_positive == 1:
                    sample_preds.append(self.idx2str.get(class_idx, self.unk))
                    pos_sample_idxs.append(sample_idx)
                    pos_class_idxs.append(class_idx)
            inv_preds.append(sample_preds)
            filtered_probs.append(probabilities[pos_sample_idxs, pos_class_idxs])

        return {
            self.predictions_key: inv_preds,
            self.probabilities_key: filtered_probs,
        }


class _SetPredict(PredictModule):
    def __init__(self, threshold):
        super().__init__()
        self.threshold = threshold

    def forward(self, inputs: Dict[str, torch.Tensor], feature_name: str) -> Dict[str, torch.Tensor]:
        logits = output_feature_utils.get_output_feature_tensor(inputs, feature_name, self.logits_key)
        probabilities = torch.sigmoid(logits)

        predictions = torch.greater_equal(probabilities, self.threshold)
        predictions = predictions.type(torch.int64)

        return {self.predictions_key: predictions, self.probabilities_key: probabilities, self.logits_key: logits}


class SetFeatureMixin(BaseFeatureMixin):
    @staticmethod
    def type():
        return SET

    @staticmethod
    def cast_column(column, backend):
        return column.astype(str)

    @staticmethod
    def get_feature_meta(
        column, preprocessing_parameters: PreprocessingConfigDict, backend, is_input_feature: bool
    ) -> FeatureMetadataDict:
        idx2str, str2idx, str2freq, max_size, _, _, _, _ = create_vocabulary(
            column,
            preprocessing_parameters["tokenizer"],
            num_most_frequent=preprocessing_parameters["most_common"],
            lowercase=preprocessing_parameters["lowercase"],
            add_special_symbols=False,
            processor=backend.df_engine,
        )
        return {
            "idx2str": idx2str,
            "str2idx": str2idx,
            "str2freq": str2freq,
            "vocab_size": len(str2idx),
            "max_set_size": max_size,
        }

    @staticmethod
    def feature_data(column, metadata, preprocessing_parameters: PreprocessingConfigDict, backend):
        def to_dense(x):
            feature_vector = set_str_to_idx(x, metadata["str2idx"], preprocessing_parameters["tokenizer"])

            set_vector = np.zeros((len(metadata["str2idx"]),))
            set_vector[feature_vector] = 1
            return set_vector.astype(np.bool_)

        return backend.df_engine.map_objects(column, to_dense)

    @staticmethod
    def add_feature_data(
        feature_config,
        input_df,
        proc_df,
        metadata,
        preprocessing_parameters: PreprocessingConfigDict,
        backend,
        skip_save_processed_input,
    ):
        proc_df[feature_config[PROC_COLUMN]] = SetFeatureMixin.feature_data(
            input_df[feature_config[COLUMN]],
            metadata[feature_config[NAME]],
            preprocessing_parameters,
            backend,
        )
        return proc_df


class SetInputFeature(SetFeatureMixin, InputFeature):
    def __init__(self, input_feature_config: SetInputFeatureConfig, encoder_obj=None, **kwargs):
        super().__init__(input_feature_config, **kwargs)

        if encoder_obj:
            self.encoder_obj = encoder_obj
        else:
            self.encoder_obj = self.initialize_encoder(input_feature_config.encoder)

    def forward(self, inputs):
        assert isinstance(inputs, torch.Tensor)
        assert inputs.dtype in [torch.bool, torch.int64, torch.float32]

        encoder_output = self.encoder_obj(inputs)

        return {"encoder_output": encoder_output}

    @property
    def input_dtype(self):
        return torch.bool

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([len(self.encoder_obj.config.vocab)])

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.encoder.vocab = feature_metadata["idx2str"]

    @staticmethod
    def get_schema_cls():
        return SetInputFeatureConfig

    @property
    def output_shape(self) -> torch.Size:
        return self.encoder_obj.output_shape

    @staticmethod
    def create_preproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _SetPreprocessing(metadata)


class SetOutputFeature(SetFeatureMixin, OutputFeature):
    def __init__(
        self,
        output_feature_config: Union[SetOutputFeatureConfig, Dict],
        output_features: Dict[str, OutputFeature],
        **kwargs,
    ):
        self.threshold = output_feature_config.threshold
        super().__init__(output_feature_config, output_features, **kwargs)
        self.decoder_obj = self.initialize_decoder(output_feature_config.decoder)
        self._setup_loss()
        self._setup_metrics()

    def logits(self, inputs, **kwargs):  # hidden
        hidden = inputs[HIDDEN]
        return self.decoder_obj(hidden)

    def loss_kwargs(self):
        return self.loss.to_dict()

    def metric_kwargs(self) -> Dict[str, Any]:
        return {"threshold": self.threshold}

    def create_predict_module(self) -> PredictModule:
        return _SetPredict(self.threshold)

    def get_prediction_set(self):
        return {PREDICTIONS, PROBABILITIES, LOGITS}

    @classmethod
    def get_output_dtype(cls):
        return torch.bool

    @property
    def input_shape(self) -> torch.Size:
        return self.decoder_obj.input_shape

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.decoder_obj.config.num_classes])

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.decoder.num_classes = feature_metadata["vocab_size"]
        if isinstance(feature_config.loss.class_weights, (list, tuple)):
            if len(feature_config.loss.class_weights) != feature_config.decoder.num_classes:
                raise ValueError(
                    "The length of class_weights ({}) is not compatible with "
                    "the number of classes ({}) for feature {}. "
                    "Check the metadata JSON file to see the classes "
                    "and their order and consider there needs to be a weight "
                    "for the <UNK> and <PAD> class too.".format(
                        len(feature_config.loss.class_weights),
                        feature_config.decoder.num_classes,
                        feature_config.name,
                    )
                )

        if isinstance(feature_config.loss.class_weights, dict):
            if feature_metadata["str2idx"].keys() != feature_config.loss.class_weights.keys():
                raise ValueError(
                    "The class_weights keys ({}) are not compatible with "
                    "the classes ({}) of feature {}. "
                    "Check the metadata JSON file to see the classes "
                    "and consider there needs to be a weight "
                    "for the <UNK> and <PAD> class too.".format(
                        feature_config.loss.class_weights.keys(),
                        feature_metadata["str2idx"].keys(),
                        feature_config.name,
                    )
                )
            else:
                class_weights = feature_config.loss.class_weights
                idx2str = feature_metadata["idx2str"]
                class_weights_list = [class_weights[s] for s in idx2str]
                feature_config.loss.class_weights = class_weights_list

    @staticmethod
    def calculate_overall_stats(predictions, targets, train_set_metadata):
        # no overall stats, just return empty dictionary
        return {}

    def postprocess_predictions(
        self,
        result,
        metadata,
    ):
        predictions_col = f"{self.feature_name}_{PREDICTIONS}"
        if predictions_col in result:

            def idx2str(pred_set):
                return [metadata["idx2str"][i] for i, pred in enumerate(pred_set) if pred]

            result[predictions_col] = result[predictions_col].map(idx2str)

        probabilities_col = f"{self.feature_name}_{PROBABILITIES}"
        if probabilities_col in result:

            def get_prob(prob_set):
                # Cast to float32 because empty np.array objects are np.float64, causing mismatch errors during saving.
                return np.array([prob for prob in prob_set if prob >= self.threshold], dtype=np.float32)

            result[probabilities_col] = result[probabilities_col].map(get_prob)

        return result

    @staticmethod
    def create_postproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _SetPostprocessing(metadata)

    @staticmethod
    def get_schema_cls():
        return SetOutputFeatureConfig



from typing import Optional


def PublicAPI(*args, **kwargs):
    """Annotation for documenting public APIs. Public APIs are classes and methods exposed to end users of Ludwig.

    If stability="stable", the APIs will remain backwards compatible across minor Ludwig releases
    (e.g., Ludwig 0.6 -> Ludwig 0.7).

    If stability="experimental", the APIs can be used by advanced users who are tolerant to and expect
    breaking changes. This will likely be seen in the case of incremental new feature development.

    Args:
        stability: One of {"stable", "experimental"}

    Examples:
        >>> from api_annotations import PublicAPI
        >>> @PublicAPI
        ... def func1(x):
        ...     return x
        >>> @PublicAPI(stability="experimental")
        ... def func2(y):
        ...     return y
    """
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        return PublicAPI(stability="stable")(args[0])

    if "stability" in kwargs:
        stability = kwargs["stability"]
        assert stability in ["stable", "experimental"], stability
    elif kwargs:
        raise ValueError(f"Unknown kwargs: {kwargs.keys()}")
    else:
        stability = "stable"

    def wrap(obj):
        if stability == "experimental":
            message = f"PublicAPI ({stability}): This API is {stability} and may change before becoming stable."
        else:
            message = "PublicAPI: This API is stable across Ludwig releases."

        _append_doc(obj, message=message)
        _mark_annotated(obj)
        return obj

    return wrap


def DeveloperAPI(*args, **kwargs):
    """Annotation for documenting developer APIs. Developer APIs are lower-level methods explicitly exposed to
    advanced Ludwig users and library developers. Their interfaces may change across minor Ludwig releases (for
    e.g., Ludwig 0.6.1 and Ludwig 0.6.2).

    Examples:
        >>> from api_annotations import DeveloperAPI
        >>> @DeveloperAPI
        ... def func(x):
        ...     return x
    """
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        return DeveloperAPI()(args[0])

    def wrap(obj):
        _append_doc(obj, message="DeveloperAPI: This API may change across minor Ludwig releases.")
        _mark_annotated(obj)
        return obj

    return wrap


def Deprecated(*args, **kwargs):
    """Annotation for documenting a deprecated API. Deprecated APIs may be removed in future releases of Ludwig
    (e.g., Ludwig 0.7 to Ludwig 0.8).

    Args:
        message: A message to help users understand the reason for the deprecation, and provide a migration path.

    Examples:
        >>> from api_annotations import Deprecated
        >>> @Deprecated
        ... def func(x):
        ...     return x
        >>> @Deprecated(message="g() is deprecated because the API is error prone. Please call h() instead.")
        ... def g(y):
        ...     return y
    """
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        return Deprecated()(args[0])

    message = "**DEPRECATED:** This API is deprecated and may be removed in a future Ludwig release."

    if "message" in kwargs:
        message += " " + kwargs["message"]
        del kwargs["message"]

    if kwargs:
        raise ValueError(f"Unknown kwargs: {kwargs.keys()}")

    def inner(obj):
        _append_doc(obj, message=message, directive="warning")
        _mark_annotated(obj)
        return obj

    return inner


def _append_doc(obj, message: str, directive: Optional[str] = None) -> str:
    """
    Args:
        message: An additional message to append to the end of docstring for a class
                 or method that uses one of the API annotations
        directive: A shorter message that provides contexts for the message and indents it.
                For example, this could be something like 'warning' or 'info'.
    """
    if not obj.__doc__:
        obj.__doc__ = ""

    obj.__doc__ = obj.__doc__.rstrip()

    indent = _get_indent(obj.__doc__)
    obj.__doc__ += "\n\n"
    if directive is not None:
        obj.__doc__ += f"{' ' * indent}.. {directive}::\n"
        obj.__doc__ += f"{' ' * (indent + 4)}{message}"
    else:
        obj.__doc__ += f"{' ' * indent}{message}"
    obj.__doc__ += f"\n{' ' * indent}"


def _mark_annotated(obj) -> None:
    # Set magic token for check_api_annotations linter.
    if hasattr(obj, "__name__"):
        obj._annotated = obj.__name__


def _is_annotated(obj) -> bool:
    # Check the magic token exists and applies to this class (not a subclass).
    return hasattr(obj, "_annotated") and obj._annotated == obj.__name__


def _get_indent(docstring: str) -> int:
    """
    Example:
        >>> def f():
        ...     '''Docstring summary.'''
        >>> f.__doc__
        'Docstring summary.'
        >>> _get_indent(f.__doc__)
        0
        >>> def g(foo):
        ...     '''Docstring summary.
        ...
        ...     Args:
        ...         foo: Does bar.
        ...     '''
        >>> g.__doc__
        'Docstring summary.\\n\\n    Args:\\n        foo: Does bar.\\n    '
        >>> _get_indent(g.__doc__)
        4
        >>> class A:
        ...     def h():
        ...         '''Docstring summary.
        ...
        ...         Returns:
        ...             None.
        ...         '''
        >>> A.h.__doc__
        'Docstring summary.\\n\\n        Returns:\\n            None.\\n        '
        >>> _get_indent(A.h.__doc__)
        8
    """
    if not docstring:
        return 0

    non_empty_lines = list(filter(bool, docstring.splitlines()))
    if len(non_empty_lines) == 1:
        # Docstring contains summary only.
        return 0

    # The docstring summary isn't indented, so check the indentation of the second non-empty line.
    return len(non_empty_lines[1]) - len(non_empty_lines[1].lstrip())



import logging
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from typing import Dict, Optional, Tuple, Union

import numpy as np
import torch

from ludwig.combiners.combiners import Combiner
from ludwig.constants import COMBINED, LOSS, NAME
from ludwig.encoders.base import Encoder
from ludwig.features.base_feature import create_passthrough_input_feature, InputFeature, OutputFeature
from ludwig.features.feature_registries import get_input_type_registry, get_output_type_registry
from ludwig.features.feature_utils import LudwigFeatureDict
from ludwig.schema.features.base import BaseInputFeatureConfig, BaseOutputFeatureConfig, FeatureCollection
from ludwig.utils.algorithms_utils import topological_sort_feature_dependencies
from ludwig.utils.metric_utils import get_scalar_from_ludwig_metric
from ludwig.utils.misc_utils import get_from_registry
from ludwig.utils.torch_utils import DEVICE, LudwigModule, reg_loss
from ludwig.utils.types import TorchDevice

logger = logging.getLogger(__name__)


class BaseModel(LudwigModule, metaclass=ABCMeta):
    """Base model for use in LudwigModule.

    Implementations of this class should implement the following methods:
    - type()
    - forward()
    """

    @staticmethod
    @abstractmethod
    def type() -> str:
        """Returns the model type."""

    def __init__(self, random_seed: int = None):
        self._random_seed = random_seed

        # TODO: with change to misc_utils.set_random_seed() this may be redundant
        #       seems to be required for test_api.py::test_api_training_determinism
        if random_seed is not None:
            torch.random.manual_seed(random_seed)

        super().__init__()

        self.input_features = LudwigFeatureDict()
        self.output_features = LudwigFeatureDict()

    @classmethod
    def build_inputs(cls, input_feature_configs: FeatureCollection[BaseInputFeatureConfig]) -> Dict[str, InputFeature]:
        """Builds and returns input features in topological order."""
        input_features = OrderedDict()
        input_features_def = topological_sort_feature_dependencies(input_feature_configs.to_list())
        for input_feature_def in input_features_def:
            input_features[input_feature_def[NAME]] = cls.build_single_input(
                getattr(input_feature_configs, input_feature_def[NAME]), input_features
            )
        return input_features

    @staticmethod
    def build_single_input(
        feature_config: BaseInputFeatureConfig, other_input_features: Optional[Dict[str, InputFeature]]
    ) -> InputFeature:
        """Builds a single input feature from the input feature definition."""
        logger.debug(f"Input {feature_config.type} feature {feature_config.name}")

        encoder_obj = None
        if feature_config.tied is not None:
            tied_input_feature_name = feature_config.tied
            if tied_input_feature_name in other_input_features:
                encoder_obj = other_input_features[tied_input_feature_name].encoder_obj

        return create_input_feature(feature_config, encoder_obj)

    @classmethod
    def build_outputs(
        cls, output_feature_configs: FeatureCollection[BaseOutputFeatureConfig], combiner: Combiner
    ) -> Dict[str, OutputFeature]:
        """Builds and returns output features in topological order."""
        output_features_def = topological_sort_feature_dependencies(output_feature_configs.to_list())
        output_features = {}

        for output_feature_def in output_features_def:
            # TODO(Justin): Check that the semantics of input_size align with what the combiner's output shape returns
            # for seq2seq.
            setattr(getattr(output_feature_configs, output_feature_def[NAME]), "input_size", combiner.output_shape[-1])
            output_features[output_feature_def[NAME]] = cls.build_single_output(
                getattr(output_feature_configs, output_feature_def[NAME]), output_features
            )
        return output_features

    @staticmethod
    def build_single_output(
        feature_config: BaseOutputFeatureConfig, output_features: Optional[Dict[str, OutputFeature]]
    ) -> OutputFeature:
        """Builds a single output feature from the output feature definition."""
        logger.debug(f"Output {feature_config.type} feature {feature_config.name}")
        output_feature_class = get_from_registry(feature_config.type, get_output_type_registry())
        output_feature_obj = output_feature_class(feature_config, output_features=output_features)
        return output_feature_obj

    def get_model_inputs(self):
        """Returns a dict of feature name -> sample model input."""
        inputs = {
            input_feature_name: input_feature.create_sample_input()
            for input_feature_name, input_feature in self.input_features.items()
        }
        return inputs

    def get_model_size(self) -> int:
        """Returns total number of parameters in model."""
        model_tensors = self.collect_weights()
        total_size = 0
        for tnsr in model_tensors:
            total_size += tnsr[1].detach().cpu().numpy().size
        return total_size

    def to_torchscript(self, device: Optional[TorchDevice] = None):
        """Converts the ECD model as a TorchScript model."""
        if device is None:
            device = DEVICE

        self.eval()
        model_inputs = self.get_model_inputs()

        model_to_script = self.to(device)
        model_inputs_to_script = {k: v.to(device) for k, v in model_inputs.items()}
        # We set strict=False to enable dict inputs and outputs.
        return torch.jit.trace(model_to_script, model_inputs_to_script, strict=False)

    def save_torchscript(self, save_path, device: Optional[TorchDevice] = None):
        """Saves the ECD model as a TorchScript model."""
        if device is None:
            device = DEVICE

        traced = self.to_torchscript(device)
        traced.save(save_path)

    @property
    def input_shape(self):
        """Returns the shape of the model's input."""
        # TODO(justin): Remove dummy implementation. Make input_shape and output_shape functions.
        return torch.Size([1, 1])

    @abstractmethod
    def forward(
        self,
        inputs: Union[
            Dict[str, torch.Tensor], Dict[str, np.ndarray], Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]
        ],
        mask=None,
    ) -> Dict[str, torch.Tensor]:
        """Forward pass of the model.

        Args:
            inputs: Inputs to the model. Can be a dictionary of input names to
                input tensors or a tuple of (inputs, targets) where inputs is
                a dictionary of input names to input tensors and targets is a
                dictionary of target names to target tensors.
            mask: A mask for the inputs.

        Returns:
            A dictionary of output {feature name}::{tensor_name} -> output tensor.
        """

    def predictions(self, inputs):
        """Returns the model's predictions for the given inputs."""
        outputs = self(inputs)
        return self.outputs_to_predictions(outputs)

    def outputs_to_predictions(self, outputs: Dict[str, torch.Tensor]) -> Dict[str, Dict[str, torch.Tensor]]:
        """Returns the model's predictions given the raw model outputs."""
        predictions = {}
        for of_name in self.output_features:
            predictions[of_name] = self.output_features.get(of_name).predictions(outputs, of_name)
        return predictions

    def evaluation_step(self, inputs, targets):
        """Predict the inputs and update evaluation metrics."""
        predictions = self.predictions(inputs)
        self.update_metrics(targets, predictions)
        return predictions

    def predict_step(self, inputs):
        """Predict the inputs."""
        return self.predictions(inputs)

    def train_loss(
        self,
        targets,
        predictions,
        regularization_type: Optional[str] = None,
        regularization_lambda: Optional[float] = None,
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Computes the training loss for the model.

        Args:
            targets: A dictionary of target names to target tensors.
            predictions: A dictionary of output names to output tensors.
            regularization_type: One of 'l1', 'l2', 'l1_l2', or None.
            regularization_lambda: The regularization lambda.

        Returns:
            A tuple of the loss tensor and a dictionary of loss for every
            output feature.
        """
        train_loss = 0
        of_train_losses = {}
        for of_name, of_obj in self.output_features.items():
            of_train_loss = of_obj.train_loss(targets[of_name], predictions, of_name)
            train_loss += of_obj.loss.weight * of_train_loss
            of_train_losses[of_name] = of_train_loss

        additional_losses = self.losses()
        if additional_losses:
            train_loss += torch.sum(torch.stack(additional_losses))  # other losses

        # Add regularization loss
        if regularization_type is not None and regularization_lambda != 0:
            train_loss += reg_loss(self, regularization_type, l1=regularization_lambda, l2=regularization_lambda)

        return train_loss, of_train_losses

    def eval_loss(self, targets, predictions):
        """Computes all evaluation losses for the model given targets and predictions.

        Args:
            targets: A dictionary of target names to target tensors.
            predictions: A dictionary of output names to output tensors.

        Returns:
            A tuple of loss values for eval losses and additional losses.
        """
        eval_loss = 0
        for of_name, of_obj in self.output_features.items():
            of_eval_loss = of_obj.eval_loss(targets[of_name], predictions[of_name])
            eval_loss += of_obj.loss.weight * of_eval_loss

        additional_loss = 0
        additional_losses = self.losses()
        if additional_losses:
            additional_loss = torch.sum(torch.stack(additional_losses))  # other losses

        return eval_loss, additional_loss

    def update_metrics(self, targets, predictions):
        """Updates the model's metrics given targets and predictions."""
        for of_name, of_obj in self.output_features.items():
            of_obj.update_metrics(targets[of_name], predictions[of_name])

        eval_loss, additional_losses = self.eval_loss(targets, predictions)
        self.eval_loss_metric.update(eval_loss)
        self.eval_additional_losses_metrics.update(additional_losses)

    def get_metrics(self) -> Dict[str, Dict[str, float]]:
        """Returns a dictionary of metrics for each output feature of the model."""
        all_of_metrics = {}
        for of_name, of_obj in self.output_features.items():
            all_of_metrics[of_name] = of_obj.get_metrics()
        all_of_metrics[COMBINED] = {
            LOSS: get_scalar_from_ludwig_metric(self.eval_loss_metric)
            + get_scalar_from_ludwig_metric(self.eval_additional_losses_metrics)
        }
        return all_of_metrics

    def reset_metrics(self):
        """Resets the model's metrics."""
        for of_obj in self.output_features.values():
            of_obj.reset_metrics()
        self.eval_loss_metric.reset()

    def collect_weights(self, tensor_names=None, **kwargs):
        """Returns named parameters filtered against `tensor_names` if not None."""
        if not tensor_names:
            return self.named_parameters()

        # Check for bad tensor names.
        weight_names = {name for name, _ in self.named_parameters()}
        for name in tensor_names:
            if name not in weight_names:
                raise ValueError(f'Requested tensor name filter "{name}" not present in the model graph')

        # Apply filter.
        tensor_set = set(tensor_names)
        return [named_param for named_param in self.named_parameters() if named_param[0] in tensor_set]

    def unskip(self):
        """Converts all skipped features into their fully encoded versions."""
        pass

    @abstractmethod
    def save(self, save_path: str):
        """Saves the model to the given path."""

    @abstractmethod
    def load(self, save_path: str):
        """Loads the model from the given path."""

    @abstractmethod
    def get_args(self):
        """Returns init arguments for constructing this model."""


def create_input_feature(feature_config: BaseInputFeatureConfig, encoder_obj: Optional[Encoder]) -> InputFeature:
    input_feature_cls = get_from_registry(feature_config.type, get_input_type_registry())
    input_feature = input_feature_cls(feature_config, encoder_obj=encoder_obj)
    if not feature_config.encoder.skip:
        return input_feature
    return create_passthrough_input_feature(input_feature, feature_config)



import logging
from typing import Callable, Tuple, Union

import torch

from ludwig.utils.torch_utils import LudwigModule

logger = logging.getLogger(__name__)


class ParameterUpdateError(Exception):
    pass


def check_module_parameters_updated(
    module: LudwigModule,
    module_input_args: Tuple,
    module_target: torch.Tensor,
    loss_function: Union[Callable, None] = None,
    max_steps: int = 1,
    learning_rate: float = 0.001,
) -> Tuple:
    """
    Reports on the number of parameters in a Ludwig component and their update status.
    Args:
        module: (LudwigModel) model to be tested.
        module_input_args: (tuple) input for model
        module_target: (Tensor) target values for computing loss and parameter updates
        loss_function: (None or Callable) Optional for module specific loss calculation
        max_steps: (int, default=1) maximum number of steps allowed to test for parameter
            updates.
        learning_rate: (float, default=0.001) learning rate for the optimizer

    Returns: Tuple(frozen_parameters, trainable_parameters, parameters_updated, not_updated)
        frozen_parameters: count of frozen parameters
        trainable_parameters: count of trainable parameters
        parameters_updated: count of trainable parameters that were updated
        not_updated: list of parameters that were not updated

    """
    # setup
    if loss_function is None:
        loss_function = torch.nn.MSELoss()
    optimizer = torch.optim.SGD(module.parameters(), lr=learning_rate)
    module.train(True)

    target_tensor = module_target

    trainable_parameter_list = []
    frozen_parameter_list = []
    parameter_updated = []
    parameters_not_updated = []
    for step in range(max_steps):
        # make pass through model
        module_output = module(*module_input_args)

        # check for any frozen parameters
        frozen_parameter_list = []
        trainable_parameter_list = []
        for p in module.named_parameters():
            if p[1].requires_grad:
                trainable_parameter_list.append(p)
            else:
                frozen_parameter_list.append(p)

        # check parameter updates only if there are some unfrozen parameters
        if len(trainable_parameter_list) > 0:
            # do update of model parameters
            optimizer.zero_grad()
            if isinstance(module_output, torch.Tensor):
                loss = loss_function(module_output, target_tensor)
            elif isinstance(module_output, dict):
                if "logits" in module_output:
                    loss = loss_function(module_output["logits"], target_tensor)
                elif "encoder_output" in module_output:
                    loss = loss_function(module_output["encoder_output"], target_tensor)
                elif "combiner_output" in module_output:
                    loss = loss_function(module_output["combiner_output"], target_tensor)
            elif isinstance(module_output, (list, tuple)):
                loss = loss_function(module_output[0], target_tensor)
            else:
                raise ValueError(f"Unexpected output type.  Module type found is {type(module_output)}")

            loss.backward()
            optimizer.step()

            # check for parameter updates
            parameter_updated = []
            # create tuple for each parameter: (parameter name, update indicator True/False)
            # parameter is deemed updated if the gradient is not None and the gradient has non-zero value
            for p in module.named_parameters():
                parameter_updated.append((p[0], (p[1].grad is not None) and (not torch.all(p[1].grad == 0))))
        else:
            parameter_updated = []

        parameters_not_updated = []
        for p in parameter_updated:
            # if not updated, record parameter name
            if not p[1]:
                parameters_not_updated.append(p[0])

    trainable_parameters = len(trainable_parameter_list)
    parameters_updated = sum(p[1] for p in parameter_updated)
    frozen_parameters = len(frozen_parameter_list)

    return frozen_parameters, trainable_parameters, parameters_updated, parameters_not_updated



from typing import Callable, Dict, List

import numpy as np
import pandas as pd
import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BINARY, CATEGORY, NAME, NUMBER, PROC_COLUMN, TYPE
from ludwig.features.feature_registries import get_input_type_registry
from ludwig.features.feature_utils import LudwigFeatureDict
from ludwig.models.base import BaseModel
from ludwig.schema.features.base import BaseInputFeatureConfig, FeatureCollection
from ludwig.types import FeatureConfigDict, TrainingSetMetadataDict
from ludwig.utils.batch_size_tuner import BatchSizeEvaluator
from ludwig.utils.dataframe_utils import from_numpy_dataset
from ludwig.utils.misc_utils import get_from_registry
from ludwig.utils.torch_utils import get_torch_device, LudwigModule

_SCALAR_TYPES = {BINARY, CATEGORY, NUMBER}


@DeveloperAPI
class Embedder(LudwigModule):
    def __init__(self, feature_configs: List[FeatureConfigDict], metadata: TrainingSetMetadataDict):
        super().__init__()

        self.input_features = LudwigFeatureDict()

        input_feature_configs = []
        for feature in feature_configs:
            feature_cls = get_from_registry(feature[TYPE], get_input_type_registry())
            feature_obj = feature_cls.get_schema_cls().from_dict(feature)
            feature_cls.update_config_with_metadata(feature_obj, metadata[feature[NAME]])

            # When running prediction or eval, we need the preprocessing to use the original pretrained
            # weights, which requires unsetting this field. In the future, we could avoid this by plumbing
            # through the saved weights and loading them dynamically after building the model.
            feature_obj.encoder.saved_weights_in_checkpoint = False

            input_feature_configs.append(feature_obj)

        feature_collection = FeatureCollection[BaseInputFeatureConfig](input_feature_configs)
        try:
            self.input_features.update(BaseModel.build_inputs(input_feature_configs=feature_collection))
        except KeyError as e:
            raise KeyError(
                f"An input feature has a name that conflicts with a class attribute of torch's ModuleDict: {e}"
            )

    def forward(self, inputs: Dict[str, torch.Tensor]):
        encoder_outputs = {}
        for input_feature_name, input_values in inputs.items():
            encoder = self.input_features.get(input_feature_name)
            encoder_output = encoder(input_values)
            encoder_outputs[input_feature_name] = encoder_output["encoder_output"]
        return encoder_outputs


@DeveloperAPI
def create_embed_batch_size_evaluator(
    features_to_encode: List[FeatureConfigDict], metadata: TrainingSetMetadataDict
) -> BatchSizeEvaluator:
    class _EmbedBatchSizeEvaluator(BatchSizeEvaluator):
        def __init__(self):
            embedder = Embedder(features_to_encode, metadata)
            self.device = get_torch_device()
            self.embedder = embedder.to(self.device)
            self.embedder.eval()

        def step(self, batch_size: int):
            inputs = {
                input_feature_name: input_feature.create_sample_input(batch_size=batch_size).to(self.device)
                for input_feature_name, input_feature in self.embedder.input_features.items()
            }
            with torch.no_grad():
                self.embedder(inputs)

    return _EmbedBatchSizeEvaluator


@DeveloperAPI
def create_embed_transform_fn(
    features_to_encode: List[FeatureConfigDict], metadata: TrainingSetMetadataDict
) -> Callable:
    class EmbedTransformFn:
        def __init__(self):
            embedder = Embedder(features_to_encode, metadata)
            self.device = get_torch_device()
            self.embedder = embedder.to(self.device)
            self.embedder.eval()

        def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
            batch = _prepare_batch(df, features_to_encode, metadata)
            name_to_proc = {i_feat.feature_name: i_feat.proc_column for i_feat in self.embedder.input_features.values()}
            inputs = {
                i_feat.feature_name: torch.from_numpy(np.array(batch[i_feat.proc_column], copy=True)).to(self.device)
                for i_feat in self.embedder.input_features.values()
            }
            with torch.no_grad():
                encoder_outputs = self.embedder(inputs)

            encoded = {name_to_proc[k]: v.detach().cpu().numpy() for k, v in encoder_outputs.items()}
            output_df = from_numpy_dataset(encoded)

            for c in output_df.columns:
                df[c] = output_df[c]

            return df

    return EmbedTransformFn


def _prepare_batch(
    df: pd.DataFrame, features: List[FeatureConfigDict], metadata: TrainingSetMetadataDict
) -> Dict[str, np.ndarray]:
    batch = {}
    for feature in features:
        c = feature[PROC_COLUMN]
        if feature[TYPE] not in _SCALAR_TYPES:
            # Ensure columns stacked instead of turned into np.array([np.array, ...], dtype=object) objects
            batch[c] = np.stack(df[c].values)
        else:
            batch[c] = df[c].to_numpy()

    for feature in features:
        c = feature[PROC_COLUMN]
        reshape = metadata.get(feature[NAME], {}).get("reshape")
        if reshape is not None:
            batch[c] = batch[c].reshape((-1, *reshape))

    return batch



from dataclasses import field

from marshmallow import fields, ValidationError

import ludwig.schema.utils as schema_utils
from ludwig.api_annotations import DeveloperAPI
from ludwig.schema.features.utils import defaults_config_registry


@DeveloperAPI
def DefaultsDataclassField(feature_type: str):
    """Custom dataclass field that when used inside a dataclass will allow the user to specify a nested default
    config for a specific feature type.

    Returns: Initialized dataclass field that converts an untyped dict with params to a defaults config.
    """

    class DefaultMarshmallowField(fields.Field):
        """Custom marshmallow field that deserializes a dict for a valid defaults config from the feature_registry
        and creates a corresponding JSON schema for external usage."""

        def _deserialize(self, value, attr, data, **kwargs):
            if value is None:
                return None
            if isinstance(value, dict):
                defaults_class = defaults_config_registry[feature_type]
                try:
                    return defaults_class.Schema().load(value)
                except (TypeError, ValidationError) as error:
                    raise ValidationError(f"Invalid params: {value}, see `{attr}` definition. Error: {error}")
            raise ValidationError(f"Invalid params: {value}")

        @staticmethod
        def _jsonschema_type_mapping():
            defaults_cls = defaults_config_registry[feature_type]
            props = schema_utils.unload_jsonschema_from_marshmallow_class(defaults_cls)["properties"]
            return {
                "type": "object",
                "properties": props,
                "additionalProperties": False,
                "title": "defaults_options",
            }

    try:
        defaults_cls = defaults_config_registry[feature_type]
        dump_default = defaults_cls.Schema().dump({})
        load_default = lambda: defaults_cls.Schema().load({})

        return field(
            metadata={
                "marshmallow_field": DefaultMarshmallowField(
                    allow_none=False,
                    dump_default=dump_default,
                    load_default=load_default,
                )
            },
            default_factory=load_default,
        )
    except Exception as e:
        raise ValidationError(f"Unsupported feature type: {feature_type}. See input_type_registry. " f"Details: {e}")



from ludwig.api_annotations import PublicAPI
from ludwig.explain.explainer import Explainer
from ludwig.explain.explanation import ExplanationsResult
from ludwig.models.gbm import GBM


@PublicAPI(stability="experimental")
class GBMExplainer(Explainer):
    def explain(self) -> ExplanationsResult:
        """Explain the model's predictions. Uses the feature importances from the model.

        # Return

        :return: ExplanationsResult containing the explanations.
            `global_explanations`: (Explanation) Aggregate explanation for the entire input data.

            `row_explanations`: (List[Explanation]) A list of explanations, one for each row in the input data. Each
            explanation contains the feature attributions for each label in the target feature's vocab.

            `expected_values`: (List[float]) of length [output feature cardinality] Expected value for each label in
            the target feature's vocab.
        """
        base_model: GBM = self.model.model
        gbm = base_model.lgbm_model
        if gbm is None:
            raise ValueError("Model has not been trained yet.")

        # Get global feature importance from the model, use it for each row in the batch.
        # TODO(travis): support local feature importance
        feat_imp = gbm.booster_.feature_importance(importance_type="gain")

        # Scale the feature importance to sum to 1.
        feat_imp = feat_imp / feat_imp.sum() if feat_imp.sum() > 0 else feat_imp

        expected_values = []
        for _ in range(self.vocab_size):
            self.global_explanation.add(base_model.input_features.keys(), feat_imp)

            for explanation in self.row_explanations:
                # Add the feature attributions to the explanation object for this row.
                explanation.add(base_model.input_features.keys(), feat_imp)

            # TODO:
            expected_values.append(0.0)

        return ExplanationsResult(self.global_explanation, self.row_explanations, expected_values)



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os
import struct
from multiprocessing.pool import ThreadPool
from typing import List, Optional

import numpy as np
import pandas as pd
import torch

from ludwig.datasets.dataset_config import DatasetConfig
from ludwig.datasets.loaders.dataset_loader import DatasetLoader
from ludwig.utils.fs_utils import makedirs

logger = logging.getLogger(__name__)
NUM_LABELS = 10


class MNISTLoader(DatasetLoader):
    def __init__(self, config: DatasetConfig, cache_dir: Optional[str] = None):
        try:
            from torchvision.io import write_png

            self.write_png = write_png
        except ImportError:
            logger.error(
                "torchvision is not installed. "
                "In order to install all image feature dependencies run "
                "pip install ludwig[image]"
            )
            raise
        super().__init__(config, cache_dir)

    def transform_files(self, file_paths: List[str]) -> List[str]:
        for dataset in ["training", "testing"]:
            labels, images = self.read_source_dataset(dataset, self.raw_dataset_dir)
            self.write_output_dataset(labels, images, os.path.join(self.raw_dataset_dir, dataset))
        return super().transform_files(file_paths)

    def load_unprocessed_dataframe(self, file_paths: List[str]) -> pd.DataFrame:
        """Load dataset files into a dataframe."""
        return self.output_training_and_test_data()

    def read_source_dataset(self, dataset="training", path="."):
        """Create a directory for training and test and extract all the images and labels to this destination.

        :args:
            dataset (str) : the label for the dataset
            path (str): the raw dataset path
        :returns:
            A tuple of the label for the image, the file array, the size and rows and columns for the image
        """
        if dataset == "training":
            fname_img = os.path.join(path, "train-images-idx3-ubyte")
            fname_lbl = os.path.join(path, "train-labels-idx1-ubyte")
        elif dataset == "testing":
            fname_img = os.path.join(path, "t10k-images-idx3-ubyte")
            fname_lbl = os.path.join(path, "t10k-labels-idx1-ubyte")
        else:
            raise ValueError("dataset must be 'testing' or 'training'")

        with open(fname_lbl, "rb") as flbl:
            struct.unpack(">II", flbl.read(8))
            lbl = np.frombuffer(flbl.read(), dtype=np.uint8)

        with open(fname_img, "rb") as fimg:
            magic_nr, size, rows, cols = struct.unpack(">IIII", fimg.read(16))
            img = np.frombuffer(fimg.read(), dtype=np.uint8)
            img = img.reshape((size, rows, cols))

        return lbl, img

    def write_output_dataset(self, labels, images, output_dir):
        """Create output directories where we write out the images.

        :args:
            labels (str) : the labels for the image
            data (np.array) : the binary array corresponding to the image
            output_dir (str) : the output directory that we need to write to
            path (str): the raw dataset path
        :returns:
            A tuple of the label for the image, the file array, the size and rows and columns for the image
        """
        # create child image output directories
        output_dirs = [os.path.join(output_dir, str(i)) for i in range(NUM_LABELS)]

        for output_dir in output_dirs:
            makedirs(output_dir, exist_ok=True)

        def write_processed_image(t):
            i, label = t
            output_filename = os.path.join(output_dirs[label], str(i) + ".png")
            torch_image = torch.from_numpy(images[i].copy()).view(1, 28, 28)
            self.write_png(torch_image, output_filename)

        # write out image data
        tasks = list(enumerate(labels))
        pool = ThreadPool(NUM_LABELS)
        pool.map(write_processed_image, tasks)
        pool.close()
        pool.join()

    def output_training_and_test_data(self):
        """Creates a combined (training and test) dataframe by iterating through all the images and labels."""
        dataframes = []
        for name in ["training", "testing"]:
            labels = []
            paths = []
            splits = []
            for i in range(NUM_LABELS):
                label_dir = f"{name}/{i}"
                img_dir = os.path.join(self.processed_dataset_dir, label_dir)
                for file in os.listdir(img_dir):
                    if file.endswith(".png"):
                        labels.append(str(i))
                        paths.append(os.path.join(label_dir, file))
                        splits.append(0 if name == "training" else 2)
            dataframes.append(pd.DataFrame({"image_path": paths, "label": labels, "split": splits}))
        return pd.concat(dataframes, ignore_index=True)



from ludwig.utils.version_transformation import VersionTransformation, VersionTransformationRegistry


def test_version_transformation_registry():
    def transform_a(config):
        config["b"] = config["a"]
        del config["a"]
        return config

    def transform_b(config):
        config["c"] = config["b"]
        del config["b"]
        return config

    def transform_e(e):
        e["g"] = e["f"]
        del e["f"]
        return e

    transformation_registry = VersionTransformationRegistry()
    transformation_registry.register(VersionTransformation(transform=transform_a, version="0.1"))
    transformation_registry.register(VersionTransformation(transform=transform_b, version="0.2"))
    transformation_registry.register(VersionTransformation(transform=transform_e, version="0.2", prefixes=["e"]))
    input_config = {"a": "a value", "e": {"f": "f_value"}}

    transformed_0_1 = transformation_registry.update_config(input_config, from_version="0.0", to_version="0.1")
    assert "a" not in transformed_0_1
    assert transformed_0_1["b"] == "a value"

    transformed_0_2 = transformation_registry.update_config(input_config, from_version="0.0", to_version="0.2")
    assert "a" not in transformed_0_2
    assert "b" not in transformed_0_2
    assert transformed_0_2["c"] == "a value"
    assert "e" in transformed_0_2
    assert "f" not in transformed_0_2["e"]
    assert transformed_0_2["e"]["g"] == "f_value"


def test_version_transformation_order():
    v1 = VersionTransformation(transform=lambda x: x, version="0.1")
    v2 = VersionTransformation(transform=lambda x: x, version="0.2")
    v3 = VersionTransformation(transform=lambda x: x, version="0.10")

    assert v1 < v2
    assert v1 < v3
    assert v2 < v3



import logging
import time
from typing import Any, List, Tuple

import numpy as np
import pandas as pd
import scipy

from ludwig.utils.types import DataFrame


def _negative_sample_user(interaction_row: np.array, neg_pos_ratio: int, extra_samples: int) -> Tuple[List[int], int]:
    """Returns a list of negative item indices for given user-item interactions.

    If there are not enough negative items, takes all of them and adds the difference to the extra_samples
    otherwise, samples with replacement.

    Params:
        interaction_row: user-item interaction row
        neg_pos_ratio: number of negative samples per positive sample
        extra_samples: number of additional samples to add to the negative sample list
    Returns:
        Tuple of list of negative item indices and number of extra samples
    """
    # Find all items that are not interacted with by the user
    neg_items = np.where(interaction_row == 0)[1]
    available_samples = len(neg_items)

    # Randomly sample negative items
    npos = interaction_row.shape[1] - len(neg_items)
    samples_required = npos * neg_pos_ratio + extra_samples
    should_sample = samples_required <= available_samples

    neg_items = np.random.choice(neg_items, samples_required, replace=False) if should_sample else neg_items

    return neg_items.tolist(), max(0, samples_required - available_samples)


def negative_sample(
    df: DataFrame,
    user_id_col: str = "customer_id",
    item_id_col: str = "article_id",
    label_col: str = "label",
    neg_pos_ratio: int = 1,
    neg_val: Any = 0,
    log_pct: int = 0,
):
    """Negative sampling for implicit feedback datasets.

    Params:
        df: DataFrame containing user-item interactions
        user_id_col: column name for user ids
        item_id_col: column name for item ids
        label_col: column name for interaction labels (e.g. 1 for positive interaction)
        n_neg: number of negative samples per positive sample
        neg_val: label value for the negative samples
        percent_print: print progress every percent_print percent. 0 to disable
    Returns:
        Input DataFrame with negative samples appended

    Source: https://petamind.com/fast-uniform-negative-sampling-for-rating-matrix/
    """
    # TODO(joppe): support out of memory negative sampling using Dask
    if not isinstance(df, pd.DataFrame):
        df = df.compute()

    # Initialize sparse COOrdinate matrix from users and items in existing interactions
    user_id_cat = df[user_id_col].astype("category").cat
    user_id_codes = user_id_cat.codes.values

    item_id_cat = df[item_id_col].astype("category").cat
    item_id_codes = item_id_cat.codes.values

    interactions_sparse = scipy.sparse.coo_matrix((df[label_col], (user_id_codes, item_id_codes)))

    # Convert to dense user-item matrix so we can iterate
    interactions_dense = interactions_sparse.todense()

    nrows = interactions_dense.shape[0]
    niter_log = int(nrows * log_pct / 100)
    start_time = time.time()

    user_indices, item_indices = [], []
    extra_samples = 0
    for user_idx, interaction_row in enumerate(interactions_dense):
        if log_pct > 0 and user_idx % niter_log == 0:
            logging.info(
                f"Negative sampling progress: {float(user_idx) * 100 / nrows:0.0f}% in {time.time() - start_time:0.2f}s"
            )

        neg_items_for_user, extra_samples = _negative_sample_user(interaction_row, neg_pos_ratio, extra_samples)

        # Add to negative user-item pairs
        item_indices += neg_items_for_user
        user_indices += [user_idx] * len(neg_items_for_user)

    negative_samples = pd.DataFrame(
        {
            # Map back to original user and item ids
            user_id_col: user_id_cat.categories[user_indices],
            item_id_col: item_id_cat.categories[item_indices],
            label_col: [neg_val] * len(item_indices),
        }
    )

    return pd.concat([df[[user_id_col, item_id_col, label_col]], negative_samples])



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Optional

import pandas as pd
from sklearn.model_selection import train_test_split

from ludwig.datasets.dataset_config import DatasetConfig
from ludwig.datasets.loaders.dataset_loader import DatasetLoader


class ForestCoverLoader(DatasetLoader):
    def __init__(self, config: DatasetConfig, cache_dir: Optional[str] = None, use_tabnet_split=True):
        super().__init__(config, cache_dir=cache_dir)
        self.use_tabnet_split = use_tabnet_split

    def transform_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        df = super().transform_dataframe(dataframe)
        # Elevation                               quantitative    meters                       Elevation in meters
        # Aspect                                  quantitative    azimuth                      Aspect in degrees azimuth
        # Slope                                   quantitative    degrees                      Slope in degrees
        # Horizontal_Distance_To_Hydrology        quantitative    meters                       Horz Dist to nearest surface water features      # noqa: E501
        # Vertical_Distance_To_Hydrology          quantitative    meters                       Vert Dist to nearest surface water features      # noqa: E501
        # Horizontal_Distance_To_Roadways         quantitative    meters                       Horz Dist to nearest roadway                     # noqa: E501
        # Hillshade_9am                           quantitative    0 to 255 index               Hillshade index at 9am, summer solstice          # noqa: E501
        # Hillshade_Noon                          quantitative    0 to 255 index               Hillshade index at noon, summer soltice          # noqa: E501
        # Hillshade_3pm                           quantitative    0 to 255 index               Hillshade index at 3pm, summer solstice          # noqa: E501
        # Horizontal_Distance_To_Fire_Points      quantitative    meters                       Horz Dist to nearest wildfire ignition points    # noqa: E501
        # Wilderness_Area (4 binary columns)      qualitative     0 (absence) or 1 (presence)  Wilderness area designation                      # noqa: E501
        # Soil_Type (40 binary columns)           qualitative     0 (absence) or 1 (presence)  Soil Type designation
        # Cover_Type (7 types)                    integer         1 to 7                       Forest Cover Type designation                    # noqa: E501

        # Map the 40 soil types to a single integer instead of 40 binary columns
        st_cols = [
            "Soil_Type_1",
            "Soil_Type_2",
            "Soil_Type_3",
            "Soil_Type_4",
            "Soil_Type_5",
            "Soil_Type_6",
            "Soil_Type_7",
            "Soil_Type_8",
            "Soil_Type_9",
            "Soil_Type_10",
            "Soil_Type_11",
            "Soil_Type_12",
            "Soil_Type_13",
            "Soil_Type_14",
            "Soil_Type_15",
            "Soil_Type_16",
            "Soil_Type_17",
            "Soil_Type_18",
            "Soil_Type_19",
            "Soil_Type_20",
            "Soil_Type_21",
            "Soil_Type_22",
            "Soil_Type_23",
            "Soil_Type_24",
            "Soil_Type_25",
            "Soil_Type_26",
            "Soil_Type_27",
            "Soil_Type_28",
            "Soil_Type_29",
            "Soil_Type_30",
            "Soil_Type_31",
            "Soil_Type_32",
            "Soil_Type_33",
            "Soil_Type_34",
            "Soil_Type_35",
            "Soil_Type_36",
            "Soil_Type_37",
            "Soil_Type_38",
            "Soil_Type_39",
            "Soil_Type_40",
        ]
        st_vals = []
        for _, row in df[st_cols].iterrows():
            st_vals.append(row.to_numpy().nonzero()[0].item(0))
        df = df.drop(columns=st_cols)
        df["Soil_Type"] = st_vals

        # Map the 4 wilderness areas to a single integer
        # instead of 4 binary columns
        wa_cols = ["Wilderness_Area_1", "Wilderness_Area_2", "Wilderness_Area_3", "Wilderness_Area_4"]
        wa_vals = []
        for _, row in df[wa_cols].iterrows():
            wa_vals.append(row.to_numpy().nonzero()[0].item(0))
        df = df.drop(columns=wa_cols)
        df["Wilderness_Area"] = wa_vals

        if not self.use_tabnet_split:
            # first 11340 records used for training data subset
            # next 3780 records used for validation data subset
            # last 565892 records used for testing data subset
            df["split"] = [0] * 11340 + [1] * 3780 + [2] * 565892
        else:
            # Split used in the tabNet paper
            # https://github.com/google-research/google-research/blob/master/tabnet/download_prepare_covertype.py
            train_val_indices, test_indices = train_test_split(range(len(df)), test_size=0.2, random_state=0)
            train_indices, val_indices = train_test_split(train_val_indices, test_size=0.2 / 0.6, random_state=0)

            df["split"] = 0
            df.loc[val_indices, "split"] = 1
            df.loc[test_indices, "split"] = 2

        return df



import logging

import pytest
import torch

from ludwig.combiners.combiners import ConcatCombiner
from ludwig.constants import CATEGORY, DECODER, NUMBER, SEQUENCE, TYPE
from ludwig.models.base import BaseModel
from ludwig.modules.reduction_modules import SequenceReducer
from ludwig.schema.model_config import ModelConfig
from ludwig.utils import output_feature_utils
from tests.integration_tests.utils import generate_output_features_with_dependencies, number_feature

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)

BATCH_SIZE = 16
SEQ_SIZE = 12
HIDDEN_SIZE = 128
OTHER_HIDDEN_SIZE = 32
OTHER_HIDDEN_SIZE2 = 64


# unit test for dependency concatenation
# tests both single and multiple dependencies
@pytest.mark.parametrize(
    "dependent_hidden_shape2",
    [
        None,
        [BATCH_SIZE, OTHER_HIDDEN_SIZE2],
        [BATCH_SIZE, SEQ_SIZE, OTHER_HIDDEN_SIZE2],
        [BATCH_SIZE, SEQ_SIZE, OTHER_HIDDEN_SIZE],
    ],
)
@pytest.mark.parametrize(
    "dependent_hidden_shape", [[BATCH_SIZE, OTHER_HIDDEN_SIZE], [BATCH_SIZE, SEQ_SIZE, OTHER_HIDDEN_SIZE]]
)
@pytest.mark.parametrize("hidden_shape", [[BATCH_SIZE, HIDDEN_SIZE], [BATCH_SIZE, SEQ_SIZE, HIDDEN_SIZE]])
@pytest.mark.parametrize(
    # todo: re-add 'attention' after further research in implication of torch
    #       migration
    "reduce_dependencies",
    ["sum", "mean", "avg", "max", "concat", "last"],
)
def test_multiple_dependencies(reduce_dependencies, hidden_shape, dependent_hidden_shape, dependent_hidden_shape2):
    # setup at least for a single dependency
    hidden_layer = torch.randn(hidden_shape, dtype=torch.float32)
    other_hidden_layer = torch.randn(dependent_hidden_shape, dtype=torch.float32)
    other_dependencies = {
        "feature_name": other_hidden_layer,
    }

    # setup dummy output feature to be root of dependency list
    num_feature_defn = number_feature()
    num_feature_defn["loss"] = {"type": "mean_squared_error"}
    num_feature_defn["dependencies"] = ["feature_name"]
    if len(dependent_hidden_shape) > 2:
        num_feature_defn["reduce_dependencies"] = reduce_dependencies

    # Based on specification calculate expected resulting hidden size for
    # with one dependencies
    if reduce_dependencies == "concat" and len(hidden_shape) == 2 and len(dependent_hidden_shape) == 3:
        expected_hidden_size = HIDDEN_SIZE + OTHER_HIDDEN_SIZE * SEQ_SIZE
    else:
        expected_hidden_size = HIDDEN_SIZE + OTHER_HIDDEN_SIZE

    # set up if multiple dependencies specified, setup second dependent feature
    if dependent_hidden_shape2:
        other_hidden_layer2 = torch.randn(dependent_hidden_shape2, dtype=torch.float32)
        other_dependencies["feature_name2"] = other_hidden_layer2
        num_feature_defn["dependencies"].append("feature_name2")
        if len(dependent_hidden_shape2) > 2:
            num_feature_defn["reduce_dependencies"] = reduce_dependencies

        # Based on specification calculate marginal increase in resulting
        # hidden size with two dependencies
        if reduce_dependencies == "concat" and len(hidden_shape) == 2 and len(dependent_hidden_shape2) == 3:
            expected_hidden_size += dependent_hidden_shape2[-1] * SEQ_SIZE
        else:
            expected_hidden_size += dependent_hidden_shape2[-1]

    # Set up dependency reducers.
    dependency_reducers = torch.nn.ModuleDict()
    for feature_name in other_dependencies.keys():
        dependency_reducers[feature_name] = SequenceReducer(reduce_mode=reduce_dependencies)

    # test dependency concatenation
    num_feature_defn["input_size"] = expected_hidden_size
    results = output_feature_utils.concat_dependencies(
        "num_feature", num_feature_defn["dependencies"], dependency_reducers, hidden_layer, other_dependencies
    )

    # confirm size of resulting concat_dependencies() call
    if len(hidden_shape) > 2:
        assert results.shape == (BATCH_SIZE, SEQ_SIZE, expected_hidden_size)
    else:
        assert results.shape == (BATCH_SIZE, expected_hidden_size)


@pytest.mark.parametrize(
    "output_feature_defs",
    [
        generate_output_features_with_dependencies("number_feature", ["category_feature"]),
        generate_output_features_with_dependencies("number_feature", ["category_feature", "sequence_feature"]),
        generate_output_features_with_dependencies("sequence_feature", ["category_feature", "number_feature"]),
    ],
)
def test_construct_output_features_with_dependencies(output_feature_defs):
    # Add keys to output_feature_defs which would have been derived from data.
    def add_data_derived_keys(output_feature_def):
        if DECODER not in output_feature_def:
            output_feature_def[DECODER] = {}
        if output_feature_def[TYPE] == CATEGORY:
            output_feature_def["num_classes"] = 2
        elif output_feature_def[TYPE] == NUMBER:
            output_feature_def[DECODER][TYPE] = "regressor"
        elif output_feature_def[TYPE] == SEQUENCE:
            output_feature_def[DECODER]["max_sequence_length"] = 5
        return output_feature_def

    output_feature_defs = [add_data_derived_keys(of) for of in output_feature_defs]
    # Gets name of output feature which has dependencies.
    dep_feature_name = [of for of in output_feature_defs if len(of.get("dependencies", [])) > 0][0]["name"]
    # Creates a dummy input feature and combiner.
    config = {
        "input_features": [number_feature()],
        "output_features": output_feature_defs,
        "combiner": {"type": "concat", "output_size": 1},
    }
    config_obj = ModelConfig.from_dict(config)
    input_features = BaseModel.build_inputs(config_obj.input_features)
    combiner = ConcatCombiner(input_features=input_features, config=config_obj.combiner)
    output_features = BaseModel.build_outputs(config_obj.output_features, combiner)
    # Gets the output feature object which has dependencies.
    feature_with_deps = output_features[dep_feature_name]
    n_dependencies = len(feature_with_deps.dependencies)
    assert n_dependencies > 0
    # Each synthetic output feature has output size 1, so total size is 1 + n_dependencies.
    assert feature_with_deps.fc_stack.input_shape == torch.Size([1 + n_dependencies])



from typing import Optional

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils
from ludwig.schema.combiners.base import BaseCombinerConfig
from ludwig.schema.metadata import COMBINER_METADATA
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class TabNetCombinerConfig(BaseCombinerConfig):
    """Parameters for tabnet combiner."""

    @staticmethod
    def module_name():
        return "TabNetCombiner"

    type: str = schema_utils.ProtectedString(
        "tabnet",
        description=COMBINER_METADATA["TabNetCombiner"]["type"].long_description,
    )

    size: int = schema_utils.PositiveInteger(
        default=32,
        description="Size of the hidden layers. `N_a` in (Arik and Pfister, 2019).",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["size"],
    )

    dropout: float = schema_utils.FloatRange(
        default=0.05,
        min=0,
        max=1,
        description="Dropout rate for the transformer block.",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["dropout"],
    )

    output_size: int = schema_utils.PositiveInteger(
        default=128,
        description="Output size of a fully connected layer. `N_d` in (Arik and Pfister, 2019).",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["output_size"],
    )

    num_steps: int = schema_utils.NonNegativeInteger(
        default=3,
        description="Number of steps / repetitions of the the attentive transformer and feature transformer "
        "computations. `N_steps` in (Arik and Pfister, 2019).",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["num_steps"],
    )

    num_total_blocks: int = schema_utils.NonNegativeInteger(
        default=4,
        description="Total number of feature transformer blocks at each step.",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["num_total_blocks"],
    )

    num_shared_blocks: int = schema_utils.NonNegativeInteger(
        default=2,
        description="Number of shared feature transformer blocks across the steps.",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["num_shared_blocks"],
    )

    relaxation_factor: float = schema_utils.FloatRange(
        default=1.5,
        description="Factor that influences how many times a feature should be used across the steps of computation. "
        "a value of 1 implies it each feature should be use once, a higher value allows for multiple "
        "usages. `gamma` in (Arik and Pfister, 2019).",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["relaxation_factor"],
    )

    bn_epsilon: float = schema_utils.FloatRange(
        default=1e-3,
        description="Epsilon to be added to the batch norm denominator.",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["bn_epsilon"],
    )

    bn_momentum: float = schema_utils.FloatRange(
        default=0.05,
        description="Momentum of the batch norm. 1 - `m_B` from the TabNet paper.",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["bn_momentum"],
    )

    bn_virtual_bs: Optional[int] = schema_utils.PositiveInteger(
        default=1024,
        allow_none=True,
        description="Size of the virtual batch size used by ghost batch norm. If null, regular batch norm is used "
        "instead. `B_v` from the TabNet paper.",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["bn_virtual_bs"],
    )

    sparsity: float = schema_utils.FloatRange(
        default=1e-4,
        description="Multiplier of the sparsity inducing loss. `lambda_sparse` in (Arik and Pfister, 2019).",
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["sparsity"],
    )

    entmax_mode: str = schema_utils.StringOptions(
        ["entmax15", "sparsemax", "constant", "adaptive"],
        default="sparsemax",
        description=(
            "Entmax is a sparse family of probability mapping which generalizes softmax and sparsemax. "
            "`entmax_mode` controls the sparsity"
        ),
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["entmax_mode"],
    )

    entmax_alpha: float = schema_utils.FloatRange(
        default=1.5,
        min=1,
        max=2,
        description=(
            "Must be a number between 1.0 and 2.0. If entmax_mode is `adaptive`, "
            "`entmax_alpha` is used as the initial value for the learnable parameter. "
            "1 corresponds to softmax, 2 is sparsemax."
        ),
        parameter_metadata=COMBINER_METADATA["TabNetCombiner"]["entmax_alpha"],
    )



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import CATEGORY, DROP_ROW, FILL_WITH_CONST, MISSING_VALUE_STRATEGY_OPTIONS, PREPROCESSING
from ludwig.schema import utils as schema_utils
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import register_preprocessor
from ludwig.schema.metadata import FEATURE_METADATA
from ludwig.schema.utils import ludwig_dataclass
from ludwig.utils import strings_utils


@DeveloperAPI
@register_preprocessor(CATEGORY)
@ludwig_dataclass
class CategoryPreprocessingConfig(BasePreprocessingConfig):
    """CategoryPreprocessingConfig is a dataclass that configures the parameters used for a category input
    feature."""

    missing_value_strategy: str = schema_utils.StringOptions(
        MISSING_VALUE_STRATEGY_OPTIONS,
        default=FILL_WITH_CONST,
        allow_none=False,
        description="What strategy to follow when there's a missing value in a category column",
        parameter_metadata=FEATURE_METADATA[CATEGORY][PREPROCESSING]["missing_value_strategy"],
    )

    fill_value: str = schema_utils.String(
        default=strings_utils.UNKNOWN_SYMBOL,
        allow_none=False,
        description=(
            "The value to replace missing values with in case the `missing_value_strategy` is `fill_with_const`"
        ),
        parameter_metadata=FEATURE_METADATA[CATEGORY][PREPROCESSING]["fill_value"],
    )

    computed_fill_value: str = schema_utils.String(
        default=strings_utils.UNKNOWN_SYMBOL,
        allow_none=False,
        description="The internally computed fill value to replace missing values with in case the "
        "missing_value_strategy is fill_with_mode or fill_with_mean",
        parameter_metadata=FEATURE_METADATA[CATEGORY][PREPROCESSING]["computed_fill_value"],
    )

    lowercase: bool = schema_utils.Boolean(
        default=False,
        description="Whether the string has to be lowercased before being handled by the tokenizer.",
        parameter_metadata=FEATURE_METADATA[CATEGORY][PREPROCESSING]["lowercase"],
    )

    most_common: int = schema_utils.PositiveInteger(
        default=10000,
        allow_none=True,
        description="The maximum number of most common tokens to be considered. if the data contains more than this "
        "amount, the most infrequent tokens will be treated as unknown.",
        parameter_metadata=FEATURE_METADATA[CATEGORY][PREPROCESSING]["most_common"],
    )


@DeveloperAPI
@register_preprocessor("category_output")
@ludwig_dataclass
class CategoryOutputPreprocessingConfig(CategoryPreprocessingConfig):
    missing_value_strategy: str = schema_utils.StringOptions(
        MISSING_VALUE_STRATEGY_OPTIONS,
        default=DROP_ROW,
        allow_none=False,
        description="What strategy to follow when there's a missing value in a category output feature",
        parameter_metadata=FEATURE_METADATA[CATEGORY][PREPROCESSING]["missing_value_strategy"],
    )

    lowercase: bool = schema_utils.Boolean(
        default=False,
        description="Whether the string has to be lowercased before being handled by the tokenizer.",
        parameter_metadata=FEATURE_METADATA[CATEGORY][PREPROCESSING]["lowercase"],
    )

    most_common: int = schema_utils.PositiveInteger(
        default=10000,
        allow_none=True,
        description="The maximum number of most common tokens to be considered. if the data contains more than this "
        "amount, the most infrequent tokens will be treated as unknown.",
        parameter_metadata=FEATURE_METADATA[CATEGORY][PREPROCESSING]["most_common"],
    )
